#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for communication with the bootloader."""

import logging
import struct
import time
from types import TracebackType
from typing import Callable, Dict, List, Optional, Sequence, Type, cast

from .commands import (
    CmdPacket,
    CmdResponse,
    CommandFlag,
    CommandTag,
    FlashReadOnceResponse,
    FlashReadResourceResponse,
    GenerateKeyBlobSelect,
    GenericResponse,
    GetPropertyResponse,
    KeyProvisioningResponse,
    KeyProvOperation,
    KeyProvUserKeyType,
    NoResponse,
    ReadMemoryResponse,
    TrustProvisioningResponse,
    TrustProvOperation,
)
from .error_codes import StatusCode, stringify_status_code
from .exceptions import (
    McuBootCommandError,
    McuBootConnectionError,
    McuBootDataAbortError,
    McuBootError,
    SPSDKError,
)
from .interfaces import MBootInterface, RawHid
from .memories import ExtMemId, ExtMemRegion, FlashRegion, MemoryRegion, RamRegion
from .properties import PropertyTag, PropertyValueBase, Version, parse_property_value

logger = logging.getLogger(__name__)


########################################################################################################################
# McuBoot Class
########################################################################################################################
class McuBoot:  # pylint: disable=too-many-public-methods
    """Class for communication with the bootloader."""

    DEFAULT_MAX_PACKET_SIZE = 32

    @property
    def status_code(self) -> int:
        """Return status code of the last operation."""
        return self._status_code

    @property
    def status_string(self) -> str:
        """Return status string."""
        return stringify_status_code(self._status_code)

    @property
    def is_opened(self) -> bool:
        """Return True if the device is open."""
        return self._device.is_opened

    def __init__(self, device: MBootInterface, cmd_exception: bool = False) -> None:
        """Initialize the McuBoot object.

        :param device: The instance of communication interface class
        :param cmd_exception: True to throw McuBootCommandError on any error;
                False to set status code only
                Note: some operation might raise McuBootCommandError is all cases

        """
        self._cmd_exception = cmd_exception
        # trick mypy to believe StatusCode really is just an integer
        self._status_code = cast(int, StatusCode.SUCCESS)
        self._device = device
        self.reopen = False
        self.enable_data_abort = False
        self._pause_point: Optional[int] = None

    def __enter__(self) -> "McuBoot":
        self.reopen = True
        self.open()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]] = None,
        exception_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        self.close()

    def _process_cmd(self, cmd_packet: CmdPacket) -> CmdResponse:
        """Process Command.

        :param cmd_packet: Command Packet
        :return: command response derived from the CmdResponse
        :raises McuBootConnectionError: Timeout Error
        :raises McuBootCommandError: Error during command execution on the target
        """
        if not self._device.is_opened:
            logger.info("TX: Device not opened")
            raise McuBootConnectionError("Device not opened")

        logger.debug(f"TX-PACKET: {cmd_packet.info()}")

        try:
            self._device.write(cmd_packet)
            response = self._device.read()
        except TimeoutError:
            self._status_code = StatusCode.NO_RESPONSE
            logger.debug("RX-PACKET: No Response, Timeout Error !")
            response = NoResponse(cmd_tag=cmd_packet.header.tag)

        assert isinstance(response, CmdResponse)
        logger.debug(f"RX-PACKET: {response.info()}")
        self._status_code = response.status  # type: ignore

        if self._cmd_exception and self._status_code != StatusCode.SUCCESS:
            raise McuBootCommandError(CommandTag.name(cmd_packet.header.tag), response.status)

        return response

    def _read_data(
        self,
        cmd_tag: int,
        length: int,
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ) -> bytes:
        """Read data from device.

        :param cmd_tag: Tag indicating the read command.
        :param length: Length of data to read
        :param progress_callback: Callback for updating the caller about the progress
        :raises McuBootConnectionError: Timeout error or a problem opening the interface
        :raises McuBootCommandError: Error during command execution on the target
        :return: Data read from the device
        """
        data = b""

        if not self._device.is_opened:
            logger.error("RX: Device not opened")
            raise McuBootConnectionError("Device not opened")
        while True:
            try:
                response = self._device.read()
            except McuBootDataAbortError as e:
                logger.error(f"RX: {e}")
                logger.info("Try increasing the timeout value")
                response = self._device.read()
            except TimeoutError:
                self._status_code = StatusCode.NO_RESPONSE
                logger.error("RX: No Response, Timeout Error !")
                response = NoResponse(cmd_tag=cmd_tag)
                break

            if isinstance(response, bytes):
                data += response
                if progress_callback:
                    progress_callback(len(data), length)

            elif isinstance(response, GenericResponse):
                logger.debug(f"RX-PACKET: {response.info()}")
                self._status_code = response.status  # type: ignore
                if response.cmd_tag == cmd_tag:
                    break

        if len(data) < length or self.status_code != StatusCode.SUCCESS:
            status_info = StatusCode.get(self._status_code, f"0x{self._status_code:08X}")
            logger.debug(f"CMD: Received {len(data)} from {length} Bytes, {status_info}")
            if self._cmd_exception:
                assert isinstance(response, CmdResponse)
                raise McuBootCommandError(CommandTag.name(cmd_tag), response.status)
        else:
            logger.info(f"CMD: Successfully Received {len(data)} from {length} Bytes")

        return data[:length] if len(data) > length else data

    def _send_data(
        self,
        cmd_tag: int,
        data: List[bytes],
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ) -> bool:
        """Send Data part of specific command.

        :param cmd_tag: Tag indicating the command
        :param data: List of data chunks to send
        :param progress_callback: Callback for updating the caller about the progress
        :raises McuBootConnectionError: Timeout error
        :raises McuBootCommandError: Error during command execution on the target
        :return: True if the operation is successful
        """
        if not self._device.is_opened:
            logger.info("TX: Device Disconnected")
            raise McuBootConnectionError("Device Disconnected !")

        total_sent = 0
        total_to_send = sum(len(chunk) for chunk in data)
        # this difference is applicable for load-image and program-aeskey commands
        expect_response = cmd_tag != CommandTag.NO_COMMAND
        self._device.allow_abort = self.enable_data_abort
        try:
            for data_chunk in data:
                self._device.write(data_chunk)
                total_sent += len(data_chunk)
                if progress_callback:
                    progress_callback(total_sent, total_to_send)
                if self._pause_point and total_sent > self._pause_point:
                    time.sleep(0.1)
                    self._pause_point = None

            if expect_response:
                response = self._device.read()
        except TimeoutError as e:
            self._status_code = StatusCode.NO_RESPONSE
            logger.error("RX: No Response, Timeout Error !")
            raise McuBootConnectionError("No Response from Device") from e
        except SPSDKError as e:
            logger.error(f"RX: {e}")
            if expect_response:
                response = self._device.read()
            else:
                self._status_code = StatusCode.SENDING_OPERATION_CONDITION_ERROR

        if expect_response:
            assert isinstance(response, CmdResponse)
            logger.debug(f"RX-PACKET: {response.info()}")
            self._status_code = response.status  # type: ignore
            if response.status != StatusCode.SUCCESS:
                status_info = StatusCode.get(self._status_code, f"0x{self._status_code:08X}")
                logger.debug(f"CMD: Send Error, {status_info}")
                if self._cmd_exception:
                    raise McuBootCommandError(CommandTag.name(cmd_tag), response.status)
                return False

        logger.info(f"CMD: Successfully Send {total_sent} out of {total_to_send} Bytes")
        return total_sent == total_to_send

    def _get_max_packet_size(self) -> int:
        """Get max packet size.

        :return int: max packet size in B
        """
        packet_size_property = None
        try:
            packet_size_property = self.get_property(prop_tag=PropertyTag.MAX_PACKET_SIZE)
        except McuBootError:
            pass
        if packet_size_property is None:
            packet_size_property = [self.DEFAULT_MAX_PACKET_SIZE]
            logger.warning(
                f"CMD: Unable to get MAX PACKET SIZE, using: {self.DEFAULT_MAX_PACKET_SIZE}"
            )
        return packet_size_property[0]

    def _split_data(self, data: bytes) -> List[bytes]:
        """Split data to send if necessary.

        :param data: Data to send
        :return: List of data splices
        """
        if not self._device.need_data_split:
            return [data]
        max_packet_size = self._get_max_packet_size()
        logger.info(f"CMD: Max Packet Size = {max_packet_size}")
        return [data[i : i + max_packet_size] for i in range(0, len(data), max_packet_size)]

    def open(self) -> None:
        """Connect to the device."""
        if not self._device.is_opened:
            logger.info(f"Connect: {self._device.info()}")
            self._device.open()

    def close(self) -> None:
        """Disconnect from the device."""
        if self._device.is_opened:
            logger.info(f"Closing: {self._device.info()}")
            self._device.close()

    def get_property_list(self) -> List[PropertyValueBase]:
        """Get a list of available properties.

        :return: List of available properties.
        :raises McuBootCommandError: Failure to read properties list
        """
        property_list: List[PropertyValueBase] = []
        for tag in PropertyTag.tags():
            try:
                values = self.get_property(tag)
            except McuBootCommandError:
                continue

            if values:
                prop = parse_property_value(tag, values)
                assert prop is not None, "Property values cannot be parsed"
                property_list.append(prop)

        self._status_code = StatusCode.SUCCESS
        if not property_list:
            self._status_code = StatusCode.FAIL
            if self._cmd_exception:
                raise McuBootCommandError("GetPropertyList", self.status_code)

        return property_list

    def _get_internal_flash(self) -> List[FlashRegion]:
        """Get information about the internal flash.

        :return: list of FlashRegion objects
        """
        index = 0
        mdata: List[FlashRegion] = []
        start_address = 0
        while True:
            try:
                values = self.get_property(PropertyTag.FLASH_START_ADDRESS, index)
                if not values:
                    break
                if index == 0:
                    start_address = values[0]
                elif start_address == values[0]:
                    break
                region_start = values[0]
                values = self.get_property(PropertyTag.FLASH_SIZE, index)
                if not values:
                    break
                region_size = values[0]
                values = self.get_property(PropertyTag.FLASH_SECTOR_SIZE, index)
                if not values:
                    break
                region_sector_size = values[0]
                mdata.append(
                    FlashRegion(
                        index=index,
                        start=region_start,
                        size=region_size,
                        sector_size=region_sector_size,
                    )
                )
                index += 1
            except McuBootCommandError:
                break

        return mdata

    def _get_internal_ram(self) -> List[RamRegion]:
        """Get information about the internal RAM.

        :return: list of RamRegion objects
        """
        index = 0
        mdata: List[RamRegion] = []
        start_address = 0
        while True:
            try:
                values = self.get_property(PropertyTag.RAM_START_ADDRESS, index)
                if not values:
                    break
                if index == 0:
                    start_address = values[0]
                elif start_address == values[0]:
                    break
                start = values[0]
                values = self.get_property(PropertyTag.RAM_SIZE, index)
                if not values:
                    break
                size = values[0]
                mdata.append(RamRegion(index=index, start=start, size=size))
                index += 1
            except McuBootCommandError:
                break

        return mdata

    def _get_ext_memories(self) -> List[ExtMemRegion]:
        """Get information about the external memories.

        :return: list of ExtMemRegion objects supported by the device
        :raises SPSDKError: If no response to get property command
        :raises SPSDKError: Other Error
        """
        ext_mem_list: List[ExtMemRegion] = []
        ext_mem_ids: Sequence[int] = ExtMemId.tags()
        try:
            values = self.get_property(PropertyTag.CURRENT_VERSION)
        except McuBootCommandError:
            values = None

        if not values and self._status_code == StatusCode.UNKNOWN_PROPERTY:
            self._status_code = StatusCode.SUCCESS
            return ext_mem_list

        if not values:
            raise SPSDKError("No response to get property command")

        if Version(values[0]) <= Version("2.0.0"):
            # old versions mboot support only Quad SPI memory
            ext_mem_ids = [ExtMemId.QUAD_SPI0]

        for mem_id in ext_mem_ids:
            try:
                values = self.get_property(PropertyTag.EXTERNAL_MEMORY_ATTRIBUTES, mem_id)
            except McuBootCommandError:
                values = None

            if not values:  # pragma: no cover  # corner-cases are currently untestable without HW
                if self._status_code == StatusCode.UNKNOWN_PROPERTY:
                    break

                if self._status_code in [
                    StatusCode.QSPI_NOT_CONFIGURED,
                    StatusCode.INVALID_ARGUMENT,
                ]:
                    continue

                if self._status_code == StatusCode.MEMORY_NOT_CONFIGURED:
                    ext_mem_list.append(ExtMemRegion(mem_id=mem_id))

                if self._status_code == StatusCode.SUCCESS:
                    raise SPSDKError("Other Error")

            else:
                ext_mem_list.append(ExtMemRegion(mem_id=mem_id, raw_values=values))
        return ext_mem_list

    def get_memory_list(self) -> dict:
        """Get list of embedded memories.

        :return: dict, with the following keys: internal_flash (optional) - list ,
                internal_ram (optional) - list, external_mems (optional) - list
        :raises McuBootCommandError: Error reading the memory list
        """
        memory_list: Dict[str, Sequence[MemoryRegion]] = {}

        # Internal FLASH
        mdata = self._get_internal_flash()
        if mdata:
            memory_list["internal_flash"] = mdata

        # Internal RAM
        ram_data = self._get_internal_ram()
        if mdata:
            memory_list["internal_ram"] = ram_data

        # External Memories
        ext_mem_list = self._get_ext_memories()
        if ext_mem_list:
            memory_list["external_mems"] = ext_mem_list

        self._status_code = StatusCode.SUCCESS
        if not memory_list:
            self._status_code = StatusCode.FAIL
            if self._cmd_exception:
                raise McuBootCommandError("GetMemoryList", self.status_code)

        return memory_list

    def flash_erase_all(self, mem_id: int = 0) -> bool:
        """Erase complete flash memory without recovering flash security section.

        :param mem_id: Memory ID
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: FlashEraseAll(mem_id={mem_id})")
        cmd_packet = CmdPacket(CommandTag.FLASH_ERASE_ALL, CommandFlag.NONE, mem_id)
        response = self._process_cmd(cmd_packet)
        return response.status == StatusCode.SUCCESS

    def flash_erase_region(self, address: int, length: int, mem_id: int = 0) -> bool:
        """Erase specified range of flash.

        :param address: Start address
        :param length: Count of bytes
        :param mem_id: Memory ID
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: FlashEraseRegion(address=0x{address:08X}, length={length}, mem_id={mem_id})"
        )
        mem_id = _clamp_down_memory_id(memory_id=mem_id)
        cmd_packet = CmdPacket(
            CommandTag.FLASH_ERASE_REGION, CommandFlag.NONE, address, length, mem_id
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def read_memory(
        self,
        address: int,
        length: int,
        mem_id: int = 0,
        progress_callback: Optional[Callable[[int, int], None]] = None,
        fast_mode: bool = False,
    ) -> Optional[bytes]:
        """Read data from MCU memory.

        :param address: Start address
        :param length: Count of bytes
        :param mem_id: Memory ID
        :param fast_mode: Fast mode for USB-HID data transfer, not reliable !!!
        :param progress_callback: Callback for updating the caller about the progress
        :return: Data read from the memory; None in case of a failure
        """
        logger.info(f"CMD: ReadMemory(address=0x{address:08X}, length={length}, mem_id={mem_id})")
        mem_id = _clamp_down_memory_id(memory_id=mem_id)

        # workaround for better USB-HID reliability
        if isinstance(self._device, RawHid) and not fast_mode:
            payload_size = self._get_max_packet_size()
            packets = length // payload_size
            remainder = length % payload_size
            if remainder:
                packets += 1

            data = b""

            for idx in range(packets):
                if idx == packets - 1 and remainder:
                    data_len = remainder
                else:
                    data_len = payload_size

                cmd_packet = CmdPacket(
                    CommandTag.READ_MEMORY,
                    CommandFlag.NONE,
                    address + idx * payload_size,
                    data_len,
                    mem_id,
                )
                cmd_response = self._process_cmd(cmd_packet)
                if cmd_response.status == StatusCode.SUCCESS:
                    data += self._read_data(CommandTag.READ_MEMORY, data_len)
                    if progress_callback:
                        progress_callback(len(data), length)
                    if self._status_code == StatusCode.NO_RESPONSE:
                        logger.warning(f"CMD: NO RESPONSE, received {len(data)}/{length} B")
                        return data
                else:
                    return b""

            return data

        cmd_packet = CmdPacket(CommandTag.READ_MEMORY, CommandFlag.NONE, address, length, mem_id)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, ReadMemoryResponse)
            return self._read_data(CommandTag.READ_MEMORY, cmd_response.length, progress_callback)
        return None

    def write_memory(
        self,
        address: int,
        data: bytes,
        mem_id: int = 0,
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ) -> bool:
        """Write data into MCU memory.

        :param address: Start address
        :param data: List of bytes
        :param progress_callback: Callback for updating the caller about the progress
        :param mem_id: Memory ID, see ExtMemId; additionally use `0` for internal memory
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: WriteMemory(address=0x{address:08X}, length={len(data)}, mem_id={mem_id})"
        )
        data_chunks = self._split_data(data=data)
        mem_id = _clamp_down_memory_id(memory_id=mem_id)
        cmd_packet = CmdPacket(
            CommandTag.WRITE_MEMORY, CommandFlag.HAS_DATA_PHASE, address, len(data), mem_id
        )
        if self._process_cmd(cmd_packet).status == StatusCode.SUCCESS:
            return self._send_data(CommandTag.WRITE_MEMORY, data_chunks, progress_callback)
        return False

    def fill_memory(self, address: int, length: int, pattern: int = 0xFFFFFFFF) -> bool:
        """Fill MCU memory with specified pattern.

        :param address: Start address (must be word aligned)
        :param length: Count of words (must be word aligned)
        :param pattern: Count of wrote bytes
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: FillMemory(address=0x{address:08X}, length={length}, pattern=0x{pattern:08X})"
        )
        cmd_packet = CmdPacket(CommandTag.FILL_MEMORY, CommandFlag.NONE, address, length, pattern)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def flash_security_disable(self, backdoor_key: bytes) -> bool:
        """Disable flash security by using of backdoor key.

        :param backdoor_key: The key value as array of 8 bytes
        :return: False in case of any problem; True otherwise
        :raises McuBootError: If the backdoor_key is not 8 bytes long
        """
        if len(backdoor_key) != 8:
            raise McuBootError("Backdoor key must by 8 bytes long")
        logger.info(f"CMD: FlashSecurityDisable(backdoor_key={backdoor_key!r})")
        key_high = backdoor_key[0:4][::-1]
        key_low = backdoor_key[4:8][::-1]
        cmd_packet = CmdPacket(
            CommandTag.FLASH_SECURITY_DISABLE, CommandFlag.NONE, data=key_high + key_low
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def get_property(self, prop_tag: PropertyTag, index: int = 0) -> Optional[List[int]]:
        """Get specified property value.

        :param prop_tag: Property TAG (see Properties Enum)
        :param index: External memory ID or internal memory region index (depends on property type)
        :return: list integers representing the property; None in case no response from device
        :raises McuBootError: If received invalid get-property response
        """
        logger.info(f"CMD: GetProperty({PropertyTag.name(prop_tag, 'UNKNOWN')!r}, index={index!r})")
        cmd_packet = CmdPacket(CommandTag.GET_PROPERTY, CommandFlag.NONE, prop_tag, index)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            if isinstance(cmd_response, GetPropertyResponse):
                return cmd_response.values
            raise McuBootError(f"Received invalid get-property response: {str(cmd_response)}")
        return None

    def set_property(self, prop_tag: PropertyTag, value: int) -> bool:
        """Set value of specified property.

        :param  prop_tag: Property TAG (see Property enumerator)
        :param  value: The value of selected property
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: SetProperty({PropertyTag.name(prop_tag)}, value=0x{value:08X})")
        cmd_packet = CmdPacket(CommandTag.SET_PROPERTY, CommandFlag.NONE, prop_tag, value)
        cmd_response = self._process_cmd(cmd_packet)
        return cmd_response.status == StatusCode.SUCCESS

    def receive_sb_file(
        self,
        data: bytes,
        progress_callback: Optional[Callable[[int, int], None]] = None,
        check_errors: bool = False,
    ) -> bool:
        """Receive SB file.

        :param  data: SB file data
        :param progress_callback: Callback for updating the caller about the progress
        :param check_errors: Check for ABORT_FRAME (and related errors) on USB interface between data packets.
            When this parameter is set to `False` significantly improves USB transfer speed (cca 20x)
            However, the final status code might be misleading (original root cause may get overridden)
            In case `receive-sb-file` fails, re-run the operation with this flag set to `True`
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: ReceiveSBfile(data_length={len(data)})")
        data_chunks = self._split_data(data=data)
        cmd_packet = CmdPacket(CommandTag.RECEIVE_SB_FILE, CommandFlag.HAS_DATA_PHASE, len(data))
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            self.enable_data_abort = check_errors
            if isinstance(self._device, RawHid):
                try:
                    # pylint: disable=import-outside-toplevel   # import only if needed to save time
                    from spsdk.sbfile.sb2.images import ImageHeaderV2

                    sb2_header = ImageHeaderV2.parse(data=data)
                    self._pause_point = sb2_header.first_boot_tag_block * 16
                except SPSDKError:
                    pass
                try:
                    # pylint: disable=import-outside-toplevel   # import only if needed to save time
                    from spsdk.sbfile.sb31.images import SecureBinary31Header

                    sb3_header = SecureBinary31Header.parse(data=data)
                    self._pause_point = sb3_header.image_total_length
                except SPSDKError:
                    pass
            result = self._send_data(CommandTag.RECEIVE_SB_FILE, data_chunks, progress_callback)
            self.enable_data_abort = False
            return result
        return False

    def execute(self, address: int, argument: int, sp: int) -> bool:  # pylint: disable=invalid-name
        """Execute program on a given address using the stack pointer.

        :param address: Jump address (must be word aligned)
        :param argument: Function arguments address
        :param sp: Stack pointer address
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: Execute(address=0x{address:08X}, argument=0x{argument:08X}, SP=0x{sp:08X})"
        )
        cmd_packet = CmdPacket(CommandTag.EXECUTE, CommandFlag.NONE, address, argument, sp)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def call(self, address: int, argument: int) -> bool:
        """Fill MCU memory with specified pattern.

        :param address: Call address (must be word aligned)
        :param argument: Function arguments address
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: Call(address=0x{address:08X}, argument=0x{argument:08X})")
        cmd_packet = CmdPacket(CommandTag.CALL, CommandFlag.NONE, address, argument)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def reset(self, timeout: int = 2000, reopen: bool = True) -> bool:
        """Reset MCU and reconnect if enabled.

        :param timeout: The maximal waiting time in [ms] for reopen connection
        :param reopen: True for reopen connection after HW reset else False
        :return: False in case of any problem; True otherwise
        :raises McuBootError: if reopen is not supported
        :raises McuBootConnectionError: Failure to reopen the device
        """
        logger.info("CMD: Reset MCU")
        cmd_packet = CmdPacket(CommandTag.RESET, CommandFlag.NONE)
        ret_val = False
        status = self._process_cmd(cmd_packet).status
        self.close()
        ret_val = True

        if status not in [StatusCode.NO_RESPONSE, StatusCode.SUCCESS]:
            ret_val = False
            if self._cmd_exception:
                raise McuBootConnectionError("Reset command failed")

        if status == StatusCode.NO_RESPONSE:
            logger.warning("Did not receive response from reset command, ignoring it")
            self._status_code = 0

        if reopen:
            if not self.reopen:
                raise McuBootError("reopen is not supported")
            time.sleep(timeout / 1000)
            try:
                self.open()
            except SPSDKError as e:
                ret_val = False
                if self._cmd_exception:
                    raise McuBootConnectionError("reopen failed") from e

        return ret_val

    def flash_erase_all_unsecure(self) -> bool:
        """Erase complete flash memory and recover flash security section.

        :return: False in case of any problem; True otherwise
        """
        logger.info("CMD: FlashEraseAllUnsecure")
        cmd_packet = CmdPacket(CommandTag.FLASH_ERASE_ALL_UNSECURE, CommandFlag.NONE)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def efuse_read_once(self, index: int) -> Optional[int]:
        """Read from MCU flash program once region.

        :param index: Start index
        :return: read value (32-bit int); None if operation failed
        """
        logger.info(f"CMD: FlashReadOnce(index={index})")
        cmd_packet = CmdPacket(CommandTag.FLASH_READ_ONCE, CommandFlag.NONE, index, 4)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, FlashReadOnceResponse)
            return cmd_response.values[0]
        return None

    def efuse_program_once(self, index: int, value: int, verify: bool = True) -> bool:
        """Write into MCU once program region (OCOTP).

        :param index: Start index
        :param value: Int value (4 bytes long)
        :param verify: Verify that data were written (by comparing value as bitmask)
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: FlashProgramOnce(index={index}, value=0x{value:X}) "
            f"with{'' if verify else 'out'} verification."
        )
        cmd_packet = CmdPacket(CommandTag.FLASH_PROGRAM_ONCE, CommandFlag.NONE, index, 4, value)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status != StatusCode.SUCCESS:
            return False
        if verify:
            read_value = self.efuse_read_once(index=index & ((1 << 24) - 1))
            if read_value is None:
                return False
            # We check only a bitmask, because OTP allows to burn individual bits separately
            # Some other bits may have been already written
            if read_value & value == value:
                return True
            # It may happen that ROM will not report error when attempting to write into locked OTP
            # In such case we substitute the original SUCCESS code with custom-made OTP_VERIFY_FAIL
            self._status_code = StatusCode.OTP_VERIFY_FAIL
            return False
        return cmd_response.status == StatusCode.SUCCESS

    def flash_read_once(self, index: int, count: int = 4) -> Optional[bytes]:
        """Read from MCU flash program once region (max 8 bytes).

        :param index: Start index
        :param count: Count of bytes
        :return: Data read; None in case of an failure
        :raises SPSDKError: When invalid count of bytes. Must be 4 or 8
        """
        if count not in (4, 8):
            raise SPSDKError("Invalid count of bytes. Must be 4 or 8")
        logger.info(f"CMD: FlashReadOnce(index={index}, bytes={count})")
        cmd_packet = CmdPacket(CommandTag.FLASH_READ_ONCE, CommandFlag.NONE, index, count)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, FlashReadOnceResponse)
            return cmd_response.data
        return None

    def flash_program_once(self, index: int, data: bytes) -> bool:
        """Write into MCU flash program once region (max 8 bytes).

        :param index: Start index
        :param data: Input data aligned to 4 or 8 bytes
        :return: False in case of any problem; True otherwise
        :raises SPSDKError: When invalid length of data. Must be aligned to 4 or 8 bytes
        """
        if len(data) not in (4, 8):
            raise SPSDKError("Invalid length of data. Must be aligned to 4 or 8 bytes")
        logger.info(f"CMD: FlashProgramOnce(index={index!r}, data={data!r})")
        cmd_packet = CmdPacket(
            CommandTag.FLASH_PROGRAM_ONCE, CommandFlag.NONE, index, len(data), data=data
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def flash_read_resource(self, address: int, length: int, option: int = 1) -> Optional[bytes]:
        """Read resource of flash module.

        :param address: Start address
        :param length: Number of bytes
        :param option: Area to be read. 0 means Flash IFR, 1 means Flash Firmware ID
        :raises McuBootError: when the length is not aligned to 4 bytes
        :return: Data from the resource; None in case of an failure
        """
        if length % 4:
            raise McuBootError("The number of bytes to read is not aligned to the 4 bytes")
        logger.info(
            f"CMD: FlashReadResource(address=0x{address:08X}, length={length}, option={option})"
        )
        cmd_packet = CmdPacket(
            CommandTag.FLASH_READ_RESOURCE, CommandFlag.NONE, address, length, option
        )
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, FlashReadResourceResponse)
            return self._read_data(CommandTag.FLASH_READ_RESOURCE, cmd_response.length)
        return None

    def configure_memory(self, address: int, mem_id: int) -> bool:
        """Configure memory.

        :param address: The address in memory where are locating configuration data
        :param mem_id: Memory ID
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: ConfigureMemory({mem_id}, address=0x{address:08X})")
        cmd_packet = CmdPacket(CommandTag.CONFIGURE_MEMORY, CommandFlag.NONE, mem_id, address)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def reliable_update(self, address: int) -> bool:
        """Reliable Update.

        :param address: Address where new the firmware is stored
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: ReliableUpdate(address=0x{address:08X})")
        cmd_packet = CmdPacket(CommandTag.RELIABLE_UPDATE, CommandFlag.NONE, address)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def generate_key_blob(
        self,
        dek_data: bytes,
        key_sel: int = GenerateKeyBlobSelect.OPTMK,
        count: int = 72,
    ) -> Optional[bytes]:
        """Generate Key Blob.

        :param dek_data: Data Encryption Key as bytes
        :param key_sel: select the BKEK used to wrap the BK (default: OPTMK/FUSES)
        :param count: Key blob count (default: 72 - AES128bit)
        :return: Key blob; None in case of an failure
        """
        logger.info(
            f"CMD: GenerateKeyBlob(dek_len={len(dek_data)}, key_sel={key_sel}, count={count})"
        )
        data_chunks = self._split_data(data=dek_data)
        cmd_response = self._process_cmd(
            CmdPacket(
                CommandTag.GENERATE_KEY_BLOB, CommandFlag.HAS_DATA_PHASE, key_sel, len(dek_data), 0
            )
        )
        if cmd_response.status != StatusCode.SUCCESS:
            return None
        if not self._send_data(CommandTag.GENERATE_KEY_BLOB, data_chunks):
            return None
        cmd_response = self._process_cmd(
            CmdPacket(CommandTag.GENERATE_KEY_BLOB, CommandFlag.NONE, key_sel, count, 1)
        )
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, ReadMemoryResponse)
            return self._read_data(CommandTag.GENERATE_KEY_BLOB, cmd_response.length)
        return None

    def kp_enroll(self) -> bool:
        """Key provisioning: Enroll Command (start PUF).

        :return: False in case of any problem; True otherwise
        """
        logger.info("CMD: [KeyProvisioning] Enroll")
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING, CommandFlag.NONE, KeyProvOperation.ENROLL
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def kp_set_intrinsic_key(self, key_type: int, key_size: int) -> bool:
        """Key provisioning: Generate Intrinsic Key.

        :param key_type: Type of the key
        :param key_size: Size of the key
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: [KeyProvisioning] SetIntrinsicKey(type={key_type}, key_size={key_size})")
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING,
            CommandFlag.NONE,
            KeyProvOperation.SET_INTRINSIC_KEY,
            key_type,
            key_size,
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def kp_write_nonvolatile(self, mem_id: int = 0) -> bool:
        """Key provisioning: Write the key to a nonvolatile memory.

        :param mem_id: The memory ID (default: 0)
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: [KeyProvisioning] WriteNonVolatileMemory(mem_id={mem_id})")
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING,
            CommandFlag.NONE,
            KeyProvOperation.WRITE_NON_VOLATILE,
            mem_id,
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def kp_read_nonvolatile(self, mem_id: int = 0) -> bool:
        """Key provisioning: Load the key from a nonvolatile memory to bootloader.

        :param mem_id: The memory ID (default: 0)
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: [KeyProvisioning] ReadNonVolatileMemory(mem_id={mem_id})")
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING,
            CommandFlag.NONE,
            KeyProvOperation.READ_NON_VOLATILE,
            mem_id,
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def kp_set_user_key(self, key_type: KeyProvUserKeyType, key_data: bytes) -> bool:
        """Key provisioning: Send the user key specified by <key_type> to bootloader.

        :param key_type: type of the user key, see enumeration for details
        :param key_data: binary content of the user key
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: [KeyProvisioning] SetUserKey(key_type={KeyProvUserKeyType.name(key_type)}, "
            f"key_len={len(key_data)})"
        )
        data_chunks = self._split_data(data=key_data)
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING,
            CommandFlag.HAS_DATA_PHASE,
            KeyProvOperation.SET_USER_KEY,
            key_type,
            len(key_data),
        )
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            return self._send_data(CommandTag.KEY_PROVISIONING, data_chunks)
        return False

    def kp_write_key_store(self, key_data: bytes) -> bool:
        """Key provisioning: Write key data into key store area.

        :param key_data: key store binary content to be written to processor
        :return: result of the operation; True means success
        """
        logger.info(f"CMD: [KeyProvisioning] WriteKeyStore(key_len={len(key_data)})")
        data_chunks = self._split_data(data=key_data)
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING,
            CommandFlag.HAS_DATA_PHASE,
            KeyProvOperation.WRITE_KEY_STORE,
            0,
            len(key_data),
        )
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            return self._send_data(CommandTag.KEY_PROVISIONING, data_chunks)
        return False

    def kp_read_key_store(self) -> Optional[bytes]:
        """Key provisioning: Read key data from key store area."""
        logger.info("CMD: [KeyProvisioning] ReadKeyStore")
        cmd_packet = CmdPacket(
            CommandTag.KEY_PROVISIONING, CommandFlag.NONE, KeyProvOperation.READ_KEY_STORE
        )
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, KeyProvisioningResponse)
            return self._read_data(CommandTag.KEY_PROVISIONING, cmd_response.length)
        return None

    def load_image(
        self, data: bytes, progress_callback: Optional[Callable[[int, int], None]] = None
    ) -> bool:
        """Load a boot image to the device.

        :param data: boot image
        :param progress_callback: Callback for updating the caller about the progress
        :return: False in case of any problem; True otherwise
        """
        logger.info(f"CMD: LoadImage(length={len(data)})")
        data_chunks = self._split_data(data)
        # there's no command in this case
        self._status_code = StatusCode.SUCCESS
        return self._send_data(CommandTag.NO_COMMAND, data_chunks, progress_callback)

    def tp_prove_genuinity(self, address: int, buffer_size: int) -> Optional[int]:
        """Start the process of proving genuinity.

        :param address: Address where to prove genuinity request (challenge) container
        :param buffer_size: Maximum size of the response package (limit 0xFFFF)
        :raises McuBootError: Invalid input parameters
        :return: True if prove_genuinity operation is successfully completed
        """
        logger.info(
            f"CMD: [TrustProvisioning] ProveGenuinity(address={hex(address)}, "
            f"buffer_size={buffer_size})"
        )
        if buffer_size > 0xFFFF:
            raise McuBootError("buffer_size must be less than 0xFFFF")
        address_msb = (address >> 32) & 0xFFFF_FFFF
        address_lsb = address & 0xFFFF_FFFF
        sentinel_cmd = _tp_sentinel_frame(
            TrustProvOperation.PROVE_GENUINITY, args=[address_msb, address_lsb, buffer_size]
        )
        cmd_packet = CmdPacket(CommandTag.TRUST_PROVISIONING, CommandFlag.NONE, data=sentinel_cmd)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:
            assert isinstance(cmd_response, TrustProvisioningResponse)
            return cmd_response.values[0]
        return None

    def tp_set_wrapped_data(self, address: int, stage: int = 0x4B, control: int = 1) -> bool:
        """Start the process of setting OEM data.

        :param address: Address where the wrapped data container on target
        :param control: 1 - use the address, 2 - use container within the firmware, defaults to 1
        :param stage: Stage of TrustProvisioning flow, defaults to 0x4B
        :return: True if set_wrapped_data operation is successfully completed
        """
        logger.info(f"CMD: [TrustProvisioning] SetWrappedData(address={hex(address)})")
        if address == 0:
            control = 2

        address_msb = (address >> 32) & 0xFFFF_FFFF
        address_lsb = address & 0xFFFF_FFFF
        stage_control = control << 8 | stage
        sentinel_cmd = _tp_sentinel_frame(
            TrustProvOperation.ISP_SET_WRAPPED_DATA, args=[stage_control, address_msb, address_lsb]
        )
        cmd_packet = CmdPacket(CommandTag.TRUST_PROVISIONING, CommandFlag.NONE, data=sentinel_cmd)
        cmd_response = self._process_cmd(cmd_packet)
        return cmd_response.status == StatusCode.SUCCESS

    def fuse_program(self, address: int, data: bytes, mem_id: int = 0) -> bool:
        """Program fuse.

        :param address: Start address
        :param data: List of bytes
        :param mem_id: Memory ID
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            f"CMD: FuseProgram(address=0x{address:08X}, length={len(data)}, mem_id={mem_id})"
        )
        data_chunks = self._split_data(data=data)
        mem_id = _clamp_down_memory_id(memory_id=mem_id)
        cmd_packet = CmdPacket(
            CommandTag.FUSE_PROGRAM, CommandFlag.HAS_DATA_PHASE, address, len(data), mem_id
        )
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:  # pragma: no cover
            # command is not supported in any device (LPC55xx family, RTxxx, RTxxxx) thus we can't measure coverage
            return self._send_data(CommandTag.FUSE_PROGRAM, data_chunks)
        return False

    def fuse_read(self, address: int, length: int, mem_id: int = 0) -> Optional[bytes]:
        """Read fuse.

        :param address: Start address
        :param length: Count of bytes
        :param mem_id: Memory ID
        :return: Data read from the fuse; None in case of a failure
        """
        logger.info(f"CMD: ReadFuse(address=0x{address:08X}, length={length}, mem_id={mem_id})")
        mem_id = _clamp_down_memory_id(memory_id=mem_id)
        cmd_packet = CmdPacket(CommandTag.FUSE_READ, CommandFlag.NONE, address, length, mem_id)
        cmd_response = self._process_cmd(cmd_packet)
        if cmd_response.status == StatusCode.SUCCESS:  # pragma: no cover
            # command is not supported in any device (LPC55xx family, RTxxx, RTxxxx) thus we can't measure coverage
            assert isinstance(cmd_response, ReadMemoryResponse)
            return self._read_data(CommandTag.FUSE_READ, cmd_response.length)
        return None

    def update_life_cycle(self, life_cycle: int) -> bool:
        """Update device life cycle.

        :param life_cycle: New life cycle value.
        :return: False in case of any problems, True otherwise.
        """
        logger.info(f"CMD: UpdateLifeCycle (life cycle=0x{life_cycle:02X})")
        cmd_packet = CmdPacket(CommandTag.UPDATE_LIFE_CYCLE, CommandFlag.NONE, life_cycle)
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def ele_message(
        self, cmdMsgAddr: int, cmdMsgCnt: int, respMsgAddr: int, respMsgCnt: int
    ) -> bool:
        """Send EdgeLock Enclave message.

        :param cmdMsgAddr: Address in RAM where is prepared the command message words
        :param cmdMsgCnt: Count of 32bits command words
        :param respMsgAddr: Address in RAM where the command store the response
        :param respMsgCnt: Count of 32bits response words

        :return: False in case of any problems, True otherwise.
        """
        logger.info(
            f"CMD: EleMessage Command (cmdMsgAddr=0x{cmdMsgAddr:08X}, cmdMsgCnt={cmdMsgCnt})"
        )
        if respMsgCnt:
            logger.info(
                f"CMD: EleMessage Response (respMsgAddr=0x{respMsgAddr:08X}, respMsgCnt={respMsgCnt})"
            )
        cmd_packet = CmdPacket(
            CommandTag.ELE_MESSAGE,
            CommandFlag.NONE,
            0,  # reserved for future use as a sub command ID or anything else
            cmdMsgAddr,
            cmdMsgCnt,
            respMsgAddr,
            respMsgCnt,
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def tp_hsm_gen_key(
        self,
        key_type: int,
        reserved: int,
        key_blob_output_addr: int,
        key_blob_output_size: int,
        ecdsa_puk_output_addr: int,
        ecdsa_puk_output_size: int,
    ) -> Optional[List[int]]:
        """Trust provisioning: OEM generate common keys.

        :param key_type: Key to generate (MFW_ISK, MFW_ENCK, GEN_SIGNK, GET_CUST_MK_SK)
        :param reserved: Reserved, must be zero
        :param key_blob_output_addr: The output buffer address where ROM writes the key blob to
        :param key_blob_output_size: The output buffer size in byte
        :param ecdsa_puk_output_addr: The output buffer address where ROM writes the public key to
        :param ecdsa_puk_output_size: The output buffer size in byte
        :return: Return byte count of the key blob + byte count of the public key from the device;
            None in case of an failure
        """
        logger.info("CMD: [TrustProvisioning] OEM generate common keys")
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.HSM_GEN_KEY,
            key_type,
            reserved,
            key_blob_output_addr,
            key_blob_output_size,
            ecdsa_puk_output_addr,
            ecdsa_puk_output_size,
        )
        cmd_response = self._process_cmd(cmd_packet)
        if isinstance(cmd_response, TrustProvisioningResponse):
            return cmd_response.values
        return None

    def tp_oem_gen_master_share(
        self,
        oem_share_input_addr: int,
        oem_share_input_size: int,
        oem_enc_share_output_addr: int,
        oem_enc_share_output_size: int,
        oem_enc_master_share_output_addr: int,
        oem_enc_master_share_output_size: int,
        oem_cust_cert_puk_output_addr: int,
        oem_cust_cert_puk_output_size: int,
    ) -> Optional[List[int]]:
        """Takes the entropy seed provided by the OEM as input.

        :param oem_share_input_addr: The input buffer address
            where the OEM Share(entropy seed) locates at
        :param oem_share_input_size: The byte count of the OEM Share
        :param oem_enc_share_output_addr: The output buffer address
            where ROM writes the Encrypted OEM Share to
        :param oem_enc_share_output_size: The output buffer size in byte
        :param oem_enc_master_share_output_addr: The output buffer address
            where ROM writes the Encrypted OEM Master Share to
        :param oem_enc_master_share_output_size: The output buffer size in byte.
        :param oem_cust_cert_puk_output_addr: The output buffer address where
            ROM writes the OEM Customer Certificate Public Key to
        :param oem_cust_cert_puk_output_size: The output buffer size in byte
        :return: Sizes of two encrypted blobs(the Encrypted OEM Share and the Encrypted OEM Master Share)
            and a public key(the OEM Customer Certificate Public Key).
        """
        logger.info("CMD: [TrustProvisioning] OEM generate master share")
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.OEM_GEN_MASTER_SHARE,
            oem_share_input_addr,
            oem_share_input_size,
            oem_enc_share_output_addr,
            oem_enc_share_output_size,
            oem_enc_master_share_output_addr,
            oem_enc_master_share_output_size,
            oem_cust_cert_puk_output_addr,
            oem_cust_cert_puk_output_size,
        )
        cmd_response = self._process_cmd(cmd_packet)
        if isinstance(cmd_response, TrustProvisioningResponse):
            return cmd_response.values
        return None

    def tp_oem_set_master_share(
        self,
        oem_share_input_addr: int,
        oem_share_input_size: int,
        oem_enc_master_share_input_addr: int,
        oem_enc_master_share_input_size: int,
    ) -> bool:
        """Takes the entropy seed and the Encrypted OEM Master Share.

        :param oem_share_input_addr: The input buffer address
            where the OEM Share(entropy seed) locates at
        :param oem_share_input_size: The byte count of the OEM Share
        :param oem_enc_master_share_input_addr: The input buffer address
            where the Encrypted OEM Master Share locates at
        :param oem_enc_master_share_input_size: The byte count of the Encrypted OEM Master Share
        :return: False in case of any problem; True otherwise
        """
        logger.info(
            "CMD: [TrustProvisioning] Takes the entropy seed and the Encrypted OEM Master Share."
        )
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.OEM_SET_MASTER_SHARE,
            oem_share_input_addr,
            oem_share_input_size,
            oem_enc_master_share_input_addr,
            oem_enc_master_share_input_size,
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def tp_oem_get_cust_cert_dice_puk(
        self,
        oem_rkth_input_addr: int,
        oem_rkth_input_size: int,
        oem_cust_cert_dice_puk_output_addr: int,
        oem_cust_cert_dice_puk_output_size: int,
    ) -> Optional[int]:
        """Creates the initial trust provisioning keys.

        :param oem_rkth_input_addr: The input buffer address where the OEM RKTH locates at
        :param oem_rkth_input_size: The byte count of the OEM RKTH
        :param oem_cust_cert_dice_puk_output_addr: The output buffer address where ROM writes the OEM Customer
            Certificate Public Key for DICE to
        :param oem_cust_cert_dice_puk_output_size: The output buffer size in byte
        :return: The byte count of the OEM Customer Certificate Public Key for DICE
        """
        logger.info("CMD: [TrustProvisioning] Creates the initial trust provisioning keys")
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.OEM_GET_CUST_CERT_DICE_PUK,
            oem_rkth_input_addr,
            oem_rkth_input_size,
            oem_cust_cert_dice_puk_output_addr,
            oem_cust_cert_dice_puk_output_size,
        )
        cmd_response = self._process_cmd(cmd_packet)
        if isinstance(cmd_response, TrustProvisioningResponse):
            return cmd_response.values[0]
        return None

    def tp_hsm_store_key(
        self,
        key_type: int,
        key_property: int,
        key_input_addr: int,
        key_input_size: int,
        key_blob_output_addr: int,
        key_blob_output_size: int,
    ) -> Optional[List[int]]:
        """Trust provisioning: OEM generate common keys.

        :param key_type: Key to generate (CKDFK, HKDFK, HMACK, CMACK, AESK, KUOK)
        :param key_property: Bit 0: Key Size, 0 for 128bit, 1 for 256bit.
            Bits 30-31: set key protection CSS mode.
        :param key_input_addr: The input buffer address where the key locates at
        :param key_input_size: The byte count of the key
        :param key_blob_output_addr: The output buffer address where ROM writes the key blob to
        :param key_blob_output_size: The output buffer size in byte
        :return: Return header of the key blob + byte count of the key blob
            (header is not included) from the device; None in case of an failure
        """
        logger.info("CMD: [TrustProvisioning] OEM generate common keys")
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.HSM_STORE_KEY,
            key_type,
            key_property,
            key_input_addr,
            key_input_size,
            key_blob_output_addr,
            key_blob_output_size,
        )
        cmd_response = self._process_cmd(cmd_packet)
        if isinstance(cmd_response, TrustProvisioningResponse):
            return cmd_response.values
        return None

    def tp_hsm_enc_blk(
        self,
        mfg_cust_mk_sk_0_blob_input_addr: int,
        mfg_cust_mk_sk_0_blob_input_size: int,
        kek_id: int,
        sb3_header_input_addr: int,
        sb3_header_input_size: int,
        block_num: int,
        block_data_addr: int,
        block_data_size: int,
    ) -> bool:
        """Trust provisioning: Encrypt the given SB3 data block.

        :param mfg_cust_mk_sk_0_blob_input_addr: The input buffer address
            where the CKDF Master Key Blob locates at
        :param mfg_cust_mk_sk_0_blob_input_size: The byte count of the CKDF Master Key Blob
        :param kek_id: The CKDF Master Key Encryption Key ID
            (0x10: NXP_CUST_KEK_INT_SK, 0x11: NXP_CUST_KEK_EXT_SK)
        :param sb3_header_input_addr: The input buffer address,
            where the SB3 Header(block0) locates at
        :param sb3_header_input_size: The byte count of the SB3 Header
        :param block_num: The index of the block. Due to SB3 Header(block 0) is always unencrypted,
            the index starts from block1
        :param block_data_addr: The buffer address where the SB3 data block locates at
        :param block_data_size: The byte count of the SB3 data block
        :return: False in case of any problem; True otherwise
        """
        logger.info("CMD: [TrustProvisioning] Encrypt the given SB3 data block")
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.HSM_ENC_BLOCK,
            mfg_cust_mk_sk_0_blob_input_addr,
            mfg_cust_mk_sk_0_blob_input_size,
            kek_id,
            sb3_header_input_addr,
            sb3_header_input_size,
            block_num,
            block_data_addr,
            block_data_size,
        )
        return self._process_cmd(cmd_packet).status == StatusCode.SUCCESS

    def tp_hsm_enc_sign(
        self,
        key_blob_input_addr: int,
        key_blob_input_size: int,
        block_data_input_addr: int,
        block_data_input_size: int,
        signature_output_addr: int,
        signature_output_size: int,
    ) -> Optional[int]:
        """Signs the given data.

        :param key_blob_input_addr: The input buffer address where signing key blob locates at
        :param key_blob_input_size: The byte count of the signing key blob
        :param block_data_input_addr: The input buffer address where the data locates at
        :param block_data_input_size: The byte count of the data
        :param signature_output_addr: The output buffer address where ROM writes the signature to
        :param signature_output_size: The output buffer size in byte
        :return: Return signature size; None in case of an failure
        """
        logger.info("CMD: [TrustProvisioning] HSM ENC SIGN")
        cmd_packet = CmdPacket(
            CommandTag.TRUST_PROVISIONING,
            CommandFlag.NONE,
            TrustProvOperation.HSM_ENC_SIGN,
            key_blob_input_addr,
            key_blob_input_size,
            block_data_input_addr,
            block_data_input_size,
            signature_output_addr,
            signature_output_size,
        )
        cmd_response = self._process_cmd(cmd_packet)
        if isinstance(cmd_response, TrustProvisioningResponse):
            return cmd_response.values[0]
        return None


####################
# Helper functions #
####################


def _tp_sentinel_frame(command: int, args: List[int], tag: int = 0x17, version: int = 0) -> bytes:
    """Prepare frame used by sentinel."""
    data = struct.pack("<4B", command, len(args), version, tag)
    for item in args:
        data += struct.pack("<I", item)
    return data


def _clamp_down_memory_id(memory_id: int) -> int:
    if memory_id > 255 or memory_id == 0:
        return memory_id
    logger.warning("Note: memoryId is not required when accessing mapped external memory")
    return 0



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing the SDP communication protocol."""

from .commands import ResponseValue
from .error_codes import StatusCode
from .exceptions import SdpCommandError, SdpConnectionError, SdpError
from .interfaces import scan_usb
from .sdp import SDP
from .sdps import SDPS

__all__ = [
    # Methods
    "scan_usb",
    # Classes
    "SDP",
    "SDPS",
    # Enums
    "ResponseValue",
    "StatusCode",
    # Errors
    "SdpError",
    "SdpCommandError",
    "SdpConnectionError",
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Internal representation of a Certificate."""

from datetime import datetime
from typing import Dict, Set

from asn1crypto import x509
from asn1crypto.keys import PublicKeyInfo
from oscrypto.asymmetric import load_public_key, rsa_pkcs1v15_verify
from oscrypto.errors import SignatureError

from .backend_internal import internal_backend
from .common import crypto_backend


class Certificate:
    """Internally used representation of a Certificate."""

    @property
    def version(self) -> str:
        """Version of the Certificate."""
        return self._cert.native["tbs_certificate"]["version"]

    @property
    def ca(self) -> bool:  # pylint: disable=invalid-name
        """Certification Authority flag."""
        return self._cert.ca

    @property
    def self_signed(self) -> str:
        """Indication whether the Certificate is self-signed.

        A unicode string of "no" or "maybe". The "maybe" result will
        be returned if the certificate issuer and subject are the same.
        If a key identifier and authority key identifier are present,
        they will need to match otherwise "no" will be returned.

        To verify is a certificate is truly self-signed, the signature
        will need to be verified. See the certvalidator package for
        one possible solution.
        """
        return self._cert.self_signed

    @property
    def self_issued(self) -> bool:
        """Is the Certificate self-issued (subject and issuer are the same)."""
        return self._cert.self_issued

    @property
    def serial_number(self) -> int:
        """Serial number of the Certificate."""
        return self._cert.serial_number

    @property
    def hash_algo(self) -> str:
        """HASH algorithm used in the Certificate."""
        return self._cert.hash_algo

    @property
    def public_key_modulus(self) -> int:
        """Modulus of the public key of the certificate."""
        return self.public_key.native["public_key"]["modulus"]

    @property
    def public_key_exponent(self) -> int:
        """Exponent of the public key of the certificate."""
        return self.public_key.native["public_key"]["public_exponent"]

    @property
    def public_key(self) -> PublicKeyInfo:
        """Public key."""
        return self._cert.public_key

    @property
    def public_key_hash(self) -> bytes:
        """32 bytes hash (SHA-256) of public key (modulus and exponent)."""
        modulus = self.public_key.native["public_key"]["modulus"]
        exponent = self.public_key.native["public_key"]["public_exponent"]
        modulus_len = (modulus.bit_length() + 7) // 8
        exponent_len = (exponent.bit_length() + 7) // 8
        return crypto_backend().hash(
            modulus.to_bytes(modulus_len, "big") + exponent.to_bytes(exponent_len, "big")
        )

    @property
    def public_key_usage(self) -> Set[str]:
        """Usage of the Certificate."""
        return self._cert.key_usage_value.native

    @property
    def signature_algo(self) -> str:
        """Certificate signature algorithm."""
        return self._cert.signature_algo

    @property
    def signature(self) -> bytes:
        """Certificate signature."""
        return self._cert.signature

    @property
    def max_path_length(self) -> int:
        """Maximum length of derived Certificate chain."""
        if not self.ca:
            return 0
        return self._cert.basic_constraints_value["path_len_constraint"].native

    @property
    def issuer(self) -> Dict[str, str]:
        """Certificate issuer."""
        return self._cert.issuer.native

    @property
    def not_valid_before(self) -> datetime:
        """Begging of the Certificate valid period."""
        return self._cert.not_valid_before

    @property
    def not_valid_after(self) -> datetime:
        """End of the Certificate valid period."""
        return self._cert.not_valid_after

    @property
    def raw_size(self) -> int:
        """Size of the Certificate data (aligned to 4 bytes)."""
        size = len(self._data)
        if size % 4:
            size += 4 - (size % 4)
        return size

    def __init__(self, certificate: bytes) -> None:
        """Initialize the Certificate object.

        :param certificate: Certificate bytes (read from file)
        """
        self._cert = x509.Certificate.load(certificate)
        self._data = certificate

    def _get_issuer_info(self, key: str, format_str: str) -> str:
        """Return selected issuer info in specified format.

        :param key: of the info to be returned
        :param format_str: formatting string, e.g. "Value: {}\n"
        :return: formatted info; empty string if the info is not available (as most of the info are optional)
        """
        value = self.issuer.get(key, None)
        if not value:
            return ""

        return format_str.format(value)

    def info(self) -> str:
        """Text information about the Certificate."""
        not_valid_before = self.not_valid_before.strftime("%d.%m.%Y (%H:%M:%S)")
        not_valid_after = self.not_valid_after.strftime("%d.%m.%Y (%H:%M:%S)")
        nfo = ""
        nfo += f"  Certification Authority:    {'YES' if self.ca else 'NO'}\n"
        if self.ca:
            nfo += f"  Max Path Length:            {self.max_path_length}\n"
        nfo += f"  Serial Number:              {hex(self.serial_number)}\n"
        nfo += f"  Validity Range:             {not_valid_before} - {not_valid_after}\n"
        nfo += f"  Signature Algorithm:        {self.signature_algo}\n"
        nfo += f"  Self Issued:                {'YES' if self.self_issued else 'NO'}\n"
        nfo += self._get_issuer_info("country_name", "  Issuer Country Name:        {}\n")
        nfo += self._get_issuer_info("state_or_province_name", "  Issuer State/Province Name: {}\n")
        nfo += self._get_issuer_info("locality_name", "  Issuer Locality Name:       {}\n")
        nfo += self._get_issuer_info("organization_name", "  Issuer Organization Name:   {}\n")
        nfo += self._get_issuer_info(
            "organizational_unit_name", "  Issuer Organ. Unit Name:    {}\n"
        )
        nfo += self._get_issuer_info("common_name", "  Issuer Common Name:         {}\n")
        nfo += self._get_issuer_info("email_address", "  Issuer Email Address:       {}\n")
        return nfo

    def export(self) -> bytes:
        """Serialized Certificate data."""
        raw_data = self._data
        if len(raw_data) % 4:
            raw_data += b"\x00" * (4 - (len(raw_data) % 4))
        return raw_data

    def verify(self, public_key_modulus: int, public_key_exponent: int) -> bool:
        """Use given public key to verify the certificate is signed.

        :param public_key_modulus: modulus of the public key to be verified
        :param public_key_exponent: exponent of the public key to be verified
        :return: True if verification pass; False otherwise
        """
        public_key = internal_backend.rsa_public_key(public_key_modulus, public_key_exponent)
        key_object = load_public_key(public_key.export_key())
        try:
            rsa_pkcs1v15_verify(
                key_object,
                self._cert["signature_value"].native,
                self._cert["tbs_certificate"].dump(),
                self.hash_algo,
            )
        except SignatureError:
            return False
        return True

    def dump(self, force: bool = False) -> bytes:
        """Encodes the value using DER.

        :param force:
            If the encoded contents already exist, clear them and regenerate
            to ensure they are in DER format instead of BER format

        :return: A byte string of the DER-encoded value
        """
        return self._cert.dump(force)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2021 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""File including helping functions."""
import functools

from spsdk import SPSDKError
from spsdk.utils.crypto.backend_internal import internal_backend
from spsdk.utils.easy_enum import Enum


class KeyDerivationMode(Enum):
    """Modes for Key derivation."""

    KDK = (1, "KDK", "Key Derivation Key mode")
    BLK = (2, "BLK", "Block Key Derivation mode")


class KeyDerivator:
    """Engine for generating derived keys."""

    def __init__(self, pck: bytes, timestamp: int, key_length: int, kdk_access_rights: int) -> None:
        """Initialize the KeyDerivator.

        :param pck: Part Common Key, base user key for all key derivations
        :param timestamp: Timestamp used for creating the KeyDerivationKey
        :param key_length: Requested key length after derivation (128/256bits)
        :param kdk_access_rights: KeyDerivationKey access rights
        """
        self.pck = pck
        self.key_length = key_length
        self.kdk_access_rights = kdk_access_rights
        self.timestamp = timestamp
        self.kdk = self._derive_kdk()

    def _derive_kdk(self) -> bytes:
        """Derive the KeyDerivationKey from PCK and timestamp."""
        return derive_kdk(self.pck, self.timestamp, self.key_length, self.kdk_access_rights)

    def get_block_key(self, block_number: int) -> bytes:
        """Derive key for particular block."""
        return derive_block_key(self.kdk, block_number, self.key_length, self.kdk_access_rights)


def derive_block_key(
    kdk: bytes, block_number: int, key_length: int, kdk_access_rights: int
) -> bytes:
    """Derive encryption AES key for given block.

    :param kdk: Key Derivation Key
    :param block_number: Block number
    :param key_length: Required key length (128/256)
    :param kdk_access_rights: Key Derivation Key access rights (0-3)
    :return: AES key for given block
    """
    return _derive_key(
        key=kdk,
        derivation_constant=block_number,
        kdk_access_rights=kdk_access_rights,
        key_length=key_length,
        mode=KeyDerivationMode.BLK,
    )


def derive_kdk(pck: bytes, timestamp: int, key_length: int, kdk_access_rights: int) -> bytes:
    """Derive the Key Derivation Key.

    :param pck: Part Common Key
    :param timestamp: Timestamp for KDK derivation
    :param key_length: Requested key length (128/256b)
    :param kdk_access_rights: KDK access rights (0-3)
    :return: Key Derivation Key
    """
    return _derive_key(
        key=pck,
        derivation_constant=timestamp,
        kdk_access_rights=kdk_access_rights,
        key_length=key_length,
        mode=KeyDerivationMode.KDK,
    )


def _derive_key(
    key: bytes,
    derivation_constant: int,
    kdk_access_rights: int,
    mode: int,
    key_length: int,
) -> bytes:
    """Derive new AES key from the provided key.

    :param key: Base (original) key
    :param derivation_constant: Derivation constant for key derivation
    :param kdk_access_rights: Key Derivation Key access rights (0-3)
    :param mode: Mode of derivation (1/2; see `KeyDerivationMode`)
    :param key_length: Requested key length (128/256b)
    :return: New (derived) AES key
    """
    # use partial to save typing later on
    derivation_data = functools.partial(
        _get_key_derivation_data,
        derivation_constant=derivation_constant,
        kdk_access_rights=kdk_access_rights,
        mode=mode,
        key_length=key_length,
    )

    result = internal_backend.cmac(data=derivation_data(iteration=1), key=key)
    if key_length == 256:
        result += internal_backend.cmac(data=derivation_data(iteration=2), key=key)
    return result


def _get_key_derivation_data(
    derivation_constant: int,
    kdk_access_rights: int,
    mode: int,
    key_length: int,
    iteration: int,
) -> bytes:
    """Generate data for AES key derivation.

    :param derivation_constant: Number for the key derivation
    :param kdk_access_rights: KeyDerivationKey access rights (0-3)
    :param mode: Mode for key derivation (1/2, see: `KeyDerivationMode`)
    :param key_length: Requested key length (128/256b)
    :param iteration: Iteration of the key derivation
    :return: Data used for key derivation
    :raises SPSDKError: Invalid mode.
    :raises SPSDKError: Invalid kdk access rights.
    :raises SPSDKError: Invalid key length.
    """
    if mode not in KeyDerivationMode.tags():
        raise SPSDKError("Invalid mode")
    if kdk_access_rights not in [0, 1, 2, 3]:
        raise SPSDKError("Invalid kdk access rights")
    if key_length not in [128, 256]:
        raise SPSDKError("Invalid key length")

    label = int.to_bytes(derivation_constant, length=12, byteorder="little")
    context = bytes(8)
    context += int.to_bytes(kdk_access_rights << 6, length=1, byteorder="big")
    context += b"\x01" if mode == KeyDerivationMode.KDK else b"\x10"
    context += bytes(1)
    key_option = 0x20 if key_length == 128 else 0x21
    context += int.to_bytes(key_option, length=1, byteorder="big")
    length = int.to_bytes(key_length, length=4, byteorder="big")
    i = int.to_bytes(iteration, length=4, byteorder="big")
    result = label + context + length + i
    return result


def add_leading_zeros(byte_data: bytes, return_size: int) -> bytes:
    """Return data with leading zeros.

    :param byte_data: Input data as bytes array
    :param return_size:
    :return: bytes
    """
    padding_size = return_size - len(byte_data)
    byte_data_with_padding = bytes("\x00" * padding_size, "utf8") + byte_data
    return byte_data_with_padding


def add_trailing_zeros(byte_data: bytes, return_size: int) -> bytes:
    """Return data with trailing zeros.

    :param byte_data: Input data as bytes array
    :param return_size:
    :return: bytes
    """
    size_of_zeros = return_size - len(byte_data)
    byte_data_with_padding = byte_data + bytes("\x00" * size_of_zeros, "utf8")
    return byte_data_with_padding



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Commands used in SmartCard."""

import logging
from struct import pack
from typing import Optional, Tuple

from spsdk import SPSDKError
from spsdk.utils.easy_enum import Enum

try:
    from smartcard.CardConnection import CardConnection
except ImportError as e:
    raise SPSDKError(
        "pyscard package is missing, please install it with pip install 'spsdk[tp]' in order to use TP"
    ) from e


from .. import SPSDKTpError

logger = logging.getLogger(__name__)


class StatusCodes(Enum):
    """SmartCard APDU general status codes."""

    SEC_COND_NOT_SATISFIED = (0x6982, "SecCondNotSatisfied", "Security condition not satisfied")
    CMD_NOT_ALLOWED_EF = (0x6986, "CmdNotAllowedEF", "Command not allowed (no current EF)")

    FILE_INVALID = (0x6983, "FileInvalid", "File contains invalid data (key, counter)")
    WRONG_DATA = (0x6A80, "WrongData", "Data field of command contains wrong data.")
    FILE_NOT_FOUND = (0x6A82, "FileNotFound", "File not found")

    SUCCESS = (0x9000, "Success", "Success")
    KEY_NOT_FOUND = (0x9008, "Key/FileNotFound", "Key/file not found")
    CMD_NOT_SUPPORTED = (0x911C, "CmdNotSupported", "Command code not supported")
    APPLET_SEL_FAILED = (0x6999, "AppletSelFail", "Applet could not be found or selected")

    ARITHMETIC_ERROR = (0x5100, "ArithmeticError", "Arithmetic error")
    APDU_ERROR = (0x5200, "ApduError", "APDU error")
    CRYPTOGRAPHIC_ERROR = (0x5300, "CryptographicError", "Cryptographic error")
    SYSTEM_ERROR = (0x5400, "SystemError", "System error")
    GENERIC_ERROR = (0x6F00, "GenericError", "Generic error")

    MAX_COUNTER_REACHED = (0x5501, "CounterReached", "Max provisioning counter reached")
    VERIFICATION_ERROR = (0x6300, "VerificationError", "Public key can't be verified")


class SmartCardAPDU:
    """Implement SmartCard APDU's operation."""

    def __init__(
        self, cla: int, ins: int, p1: int, p2: int, data: Optional[bytes] = None, le: int = 0
    ) -> None:
        """Simple APDU transfer descriptor.

        :param cla: Class of instruction - indicates the structure and format for a category of
                    command and response APDUs
        :param ins: Instruction code: specifies the instruction of the command
        :param p1: Instruction parameter 1 - further provide qualifications to the instruction
        :param p2: Instruction parameter 2 - further provide qualifications to the instruction
        :param data: A sequence of bytes in the data field of the command
        :param le: Maximum of bytes expected in the data field of the response to the command
        """
        assert cla in [0x00, 0x80]
        assert ins < 256
        assert p1 < 256
        assert p2 < 256
        assert le < 65535
        if data:
            assert len(data) < 65535

        self.cla = cla
        self.ins = ins
        self.p1 = p1
        self.p2 = p2
        self.lc = len(data) if data else 0
        self.data = data
        self.le = le
        self.is_extended = self.lc > 255 or self.le > 255
        self.status = 0
        self.response = bytes()

    def get_command(self) -> bytes:
        """Get binary command data."""
        fmt = "<BBBB"
        ret = pack(fmt, self.cla, self.ins, self.p1, self.p2)
        if self.data:
            ret += self.lc.to_bytes(length=3 if self.is_extended else 1, byteorder="big")
            ret += self.data
        if self.le:
            ret += self.le.to_bytes(length=2 if self.is_extended else 1, byteorder="big")
        return ret

    @classmethod
    def get_status_description(cls, code: int) -> str:
        """Return text description of status code."""
        desc = StatusCodes.desc(code)
        if not desc:
            desc = StatusCodes.desc(code & 0xFF00, "Unknown")
        return f"{hex(code)}: {desc}"

    @classmethod
    def format_error_message(cls, status_code: int, extra_message: Optional[str] = None) -> str:
        """Format error message.

        :param status_code: Status code of the operation.
        :param extra_message: Extra prefix in the error message, defaults to None
        :return: String representing the error.
        """
        message = extra_message or ""
        message += ": " if message else ""
        message += cls.get_status_description(status_code)
        return message

    def transmit(self, connection: CardConnection, extra_message: Optional[str] = None) -> bytes:
        """Transmit the command data using `connection`.

        :param connection: Connection to use for transmission
        :param extra_message: Extra message in case of an error, defaults to None
        :raises SPSDKTpError: If an error occurs (status != 0x9000)
        :return: Response data without status word
        """
        command_data = list(self.get_command())
        response, sw1, sw2 = connection.transmit(command_data)
        self.response = bytes(response)
        self.status = (sw1 << 8) + sw2
        if self.status != StatusCodes.SUCCESS:
            raise SPSDKTpError(self.format_error_message(self.status, extra_message))
        return self.response


#################################
# TRUST-PROVISIONING OPERATIONS #
#################################


class SetProvisioningItem(SmartCardAPDU):
    """Set provisioning data item."""

    def __init__(self, prov_item: int, data: bytes) -> None:
        """Set provisioning data item."""
        p1, p2 = int_to_p1p2(prov_item)
        super().__init__(cla=0x80, ins=0xDA, p1=p1, p2=p2, data=data)


class GetChallenge(SmartCardAPDU):
    """Get TP Challenge."""

    def __init__(self) -> None:
        """Get TP Challenge."""
        super().__init__(cla=0x00, ins=0x84, p1=0x00, p2=0x00)


class ProcessTPResponse(SmartCardAPDU):
    """Process the TP Response. Obtain the WRAPPED DATA."""

    def __init__(self, tp_response: bytes) -> None:
        """Process the TP Response. Obtain the WRAPPED DATA."""
        super().__init__(cla=0x00, ins=0x89, p1=0x00, p2=0x00, data=tp_response)


class GetProductionCounter(SmartCardAPDU):
    """Get Current value of the production counter."""

    def __init__(self) -> None:
        """Get Current value of the production counter."""
        super().__init__(cla=0x00, ins=0xCA, p1=0x01, p2=0x17)

    @staticmethod
    def format(response: bytes) -> int:
        """Format the `transmit` response into `int`."""
        return int.from_bytes(response, byteorder="big")


class GetProductionRemainder(SmartCardAPDU):
    """Get Current value of the production counter."""

    def __init__(self) -> None:
        """Get Current value of the production counter."""
        super().__init__(cla=0x00, ins=0xCA, p1=0x01, p2=0x18)

    @staticmethod
    def format(response: bytes) -> int:
        """Format the `transmit` response into `int`."""
        return int.from_bytes(response, byteorder="big")


class Select(SmartCardAPDU):
    """APDU Select command."""

    def __init__(self, applet: str) -> None:
        """Select an applet with given name."""
        applet_id = applet.encode("utf-8")
        super().__init__(cla=0x00, ins=0xA4, p1=0x04, p2=0x00, data=applet_id)


#######################
# GET DATA OPERATIONS #
#######################


class GetAppletName(SmartCardAPDU):
    """Get the applet name."""

    def __init__(self) -> None:
        """Get the applet name."""
        super().__init__(cla=0x00, ins=0xCA, p1=0x01, p2=0x00)

    @staticmethod
    def format(response: bytes) -> str:
        """Format the `transmit` response into `str`."""
        return bytes(response).decode("utf-8")


class GetSerialNumber(SmartCardAPDU):
    """Get Serial Number."""

    def __init__(self) -> None:
        """Get Serial Number."""
        super().__init__(cla=0x00, ins=0xCA, p1=0x01, p2=0x14)

    @staticmethod
    def format(response: bytes) -> int:
        """Format the `transmit` response into `int`."""
        return int.from_bytes(response, byteorder="big")


class GetAppletVersion(SmartCardAPDU):
    """Get Applet version command."""

    def __init__(self) -> None:
        """Get Version command."""
        super().__init__(cla=0x00, ins=0xCA, p1=0x01, p2=0x16)

    @staticmethod
    def format(response: bytes) -> str:
        """Format the `transmit` response into `str`."""
        return ".".join(str(i) for i in response[:3])


class GetFreeMemory(SmartCardAPDU):
    """Get free memory."""

    def __init__(self, memory_type: int) -> None:
        """Get free memory of a given type.

        :param memory_type: 1 - NVM, 2 - COR, 3 - COD
        """
        super().__init__(cla=0x00, ins=0xCA, p1=0x02, p2=memory_type)

    @staticmethod
    def format(response: bytes) -> int:
        """Format the `transmit` response into `int`."""
        return int.from_bytes(response, byteorder="big")


class Echo(SmartCardAPDU):
    """Send data and receive it back."""

    def __init__(self, data: Optional[bytes] = None) -> None:
        """Send `data` and receive it back."""
        super().__init__(cla=0x80, ins=0x00, p1=0x00, p2=0x00, data=data)


#####################
# BUFFER OPERATIONS #
#####################


class ResizeNVMBuffer(SmartCardAPDU):
    """Resize NVM buffer."""

    def __init__(self, new_length: int) -> None:
        """Resize NVM buffer to `new_length`."""
        length_data = new_length.to_bytes(length=2, byteorder="big")
        super().__init__(cla=0x80, ins=0x84, p1=0x00, p2=0x00, data=length_data)


class DeleteNVMBuffer(SmartCardAPDU):
    """Delete NVM Buffer. Needs to be called before ResizeNVMBuffer."""

    def __init__(self) -> None:
        """Delete NVM Buffer."""
        super().__init__(cla=0x80, ins=0x84, p1=0x00, p2=0x00, data=bytes(2))


class ResizeTransientBuffer(SmartCardAPDU):
    """Resize NVM buffer."""

    def __init__(self, new_length: int) -> None:
        """Resize NVM buffer to `new_length`."""
        length_data = new_length.to_bytes(length=2, byteorder="big")
        super().__init__(cla=0x80, ins=0x84, p1=0x00, p2=0x01, data=length_data)


class DeleteTransientBuffer(SmartCardAPDU):
    """Delete Transient Buffer. Needs to be called before ResizeTransientBuffer."""

    def __init__(self) -> None:
        """Delete Transient Buffer."""
        super().__init__(cla=0x80, ins=0x84, p1=0x00, p2=0x01, data=bytes(2))


##########################
# FILE SYSTEM OPERATIONS #
##########################


class CreateFileSystem(SmartCardAPDU):
    """Create File System on applet. Turns applet into PERSONALIZATION mode."""

    def __init__(self, objects_count: int = 15) -> None:
        """Create File System on applet."""
        super().__init__(cla=0x80, ins=0xE1, p1=00, p2=objects_count)


class DeleteFileSystem(SmartCardAPDU):
    """Delete the whole filesystem on applet. Turns applet into PRE-PERSONALIZATION mode."""

    def __init__(self) -> None:
        """Delete the whole filesystem on applet."""
        super().__init__(cla=0x80, ins=0xE4, p1=0x80, p2=0x00)


class DeleteFile(SmartCardAPDU):
    """Delete single file."""

    def __init__(self, file_index: int) -> None:
        """Delete single file with index `file_index`."""
        p1, p2 = int_to_p1p2(file_index)
        super().__init__(cla=0x00, ins=0xE4, p1=p1, p2=p2)


class FinalizeFileSystem(SmartCardAPDU):
    """Finalizes the file system personalization. Turns applet into OPERATIONAL mode."""

    def __init__(self) -> None:
        """Finalizes the file system personalization."""
        super().__init__(cla=0x80, ins=0x44, p1=0x00, p2=0x00)


def int_to_p1p2(param: int) -> Tuple[int, int]:
    """Converts integer into the bytes (p1, p2)."""
    p1 = (param >> 8) & 0xFF
    p2 = param & 0xFF
    return p1, p2


def p1p2_to_int(p1: int, p2: int) -> int:
    """Converts two bytes (p1, p2) into integer."""
    param = p1 & 0xFF << 8 + p2 & 0xFF
    return param



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2021 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Exceptions used in the MBoot module."""

from spsdk import SPSDKError

from .error_codes import StatusCode

########################################################################################################################
# McuBoot Exceptions
########################################################################################################################


class McuBootError(SPSDKError):
    """MBoot Module: Base Exception."""

    fmt = "MBoot: {description}"


class McuBootCommandError(McuBootError):
    """MBoot Module: Command Exception."""

    fmt = "MBoot: {cmd_name} interrupted -> {description}"

    def __init__(self, cmd: str, value: int) -> None:
        """Initialize the Command Error exception.

        :param cmd: Name of the command causing the exception
        :param value: Response value causing the exception
        """
        super().__init__()
        self.cmd_name = cmd
        self.error_value = value
        self.description = StatusCode.desc(value, f"Unknown Error 0x{value:08X}")

    def __str__(self) -> str:
        return self.fmt.format(cmd_name=self.cmd_name, description=self.description)


class McuBootDataAbortError(McuBootError):
    """MBoot Module: Data phase aborted by sender."""

    fmt = "Mboot: Data aborted by sender"


class McuBootConnectionError(McuBootError):
    """MBoot Module: Connection Exception."""

    fmt = "MBoot: Connection issue -> {description}"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for key generation and saving keys to file (RSA and ECC)."""

from enum import Enum
from typing import Optional

from spsdk.crypto import (
    EllipticCurvePrivateKeyWithSerialization,
    EllipticCurvePublicKey,
    Encoding,
    RSAPrivateKeyWithSerialization,
    RSAPublicKey,
    RSAPublicNumbers,
    default_backend,
    ec,
    rsa,
    serialization,
)
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.misc import write_file


class CurveName(str, Enum):
    """Supported ecc key types."""

    PRIME192V1 = "prime192v1"
    PRIME256V1 = "prime256v1"
    SECP192R1 = "secp192r1"
    SECP224R1 = "secp224r1"
    SECP256R1 = "secp256r1"
    SECP384R1 = "secp384r1"
    SECP521R1 = "secp521r1"
    SECP256K1 = "secp256k1"
    SECT163K1 = "sect163k1"
    SECT233K1 = "sect233k1"
    SECT283K1 = "sect283k1"
    SECT409K1 = "sect409k1"
    SECT571K1 = "sect571k1"
    SECT163R2 = "sect163r2"
    SECT233R1 = "sect233r1"
    SECT283R1 = "sect283r1"
    SECT409R1 = "sect409r1"
    SECT571R1 = "sect571r1"
    BrainpoolP256R1 = "brainpoolP256r1"  # pylint: disable=invalid-name
    BrainpoolP384R1 = "brainpoolP384r1"  # pylint: disable=invalid-name
    BrainpoolP512R1 = "brainpoolP512r1"  # pylint: disable=invalid-name


def get_ec_curve_object(name: str) -> ec.EllipticCurve:
    """Get the EC curve object by its name.

    :param name: Name of EC curve.
    :return: EC curve object.
    :raises SPSDKValueError: Invalid EC curve name.
    """
    # pylint: disable=protected-access
    for key_object in ec._CURVE_TYPES:  # type: ignore
        if key_object.lower() == name.lower():
            # pylint: disable=protected-access
            return ec._CURVE_TYPES[key_object]()  # type: ignore

    raise SPSDKValueError(f"The EC curve with name '{name}' is not supported.")


def generate_rsa_private_key(
    key_size: int = 2048, exponent: int = 65537
) -> RSAPrivateKeyWithSerialization:
    """Generate RSA private key.

    :param key_size: key size in bits; must be >= 512
    :param exponent: public exponent; must be >= 3 and odd
    :return: RSA private key with serialization
    """
    return rsa.generate_private_key(
        backend=default_backend(), public_exponent=exponent, key_size=key_size
    )


def generate_rsa_public_key(
    private_key: RSAPrivateKeyWithSerialization,
) -> RSAPublicKey:
    """Generate RSA public key.

    :param private_key: private key used for public key generation
    :return: RSA public key
    """
    return private_key.public_key()


def save_rsa_private_key(
    private_key: RSAPrivateKeyWithSerialization,
    file_path: str,
    password: Optional[str] = None,
    encoding: Encoding = Encoding.PEM,
) -> None:
    """Save the RSA private key to the given file.

    :param private_key: RSA private key to be saved
    :param file_path: path to the file, where the key will be stored
    :param password: password to private key; None to store without password
    :param encoding: encoding type, default is PEM
    """
    if password:
        if isinstance(password, str):
            password_bytes = password.encode("utf-8")
        else:
            password_bytes = password
    enc = (
        serialization.BestAvailableEncryption(password=password_bytes)
        if password
        else serialization.NoEncryption()
    )
    pem_data = private_key.private_bytes(encoding, serialization.PrivateFormat.PKCS8, enc)
    write_file(pem_data, file_path, mode="wb")


def save_rsa_public_key(
    public_key: RSAPublicKey, file_path: str, encoding: Encoding = Encoding.PEM
) -> None:
    """Save the RSA public key to the file.

    :param public_key: public key to be saved
    :param file_path: path to the file, where the key will be stored
    :param encoding: encoding type, default is PEM
    """
    pem_data = public_key.public_bytes(encoding, serialization.PublicFormat.PKCS1)
    write_file(pem_data, file_path, mode="wb")


# pylint: disable=invalid-name  # e,n are commonly used symbols for exponent and modulus
def recreate_rsa_public_key(e: int, n: int) -> RSAPublicKey:
    """Recreate RSA public key from Exponent and modulus.

    :param e: Exponent of RSA key.
    :param n: Modulus of RSA key.
    :return: RSA public key.
    """
    public_numbers = RSAPublicNumbers(e=e, n=n)
    return public_numbers.public_key()


def generate_ecc_private_key(curve_name: str) -> EllipticCurvePrivateKeyWithSerialization:
    """Generate ECC private key.

    :param curve_name: name of curve
    :return: ECC private key
    """
    curve_obj = get_ec_curve_object(curve_name)
    return ec.generate_private_key(curve_obj, default_backend())  # type: ignore


def generate_ecc_public_key(
    private_key: EllipticCurvePrivateKeyWithSerialization,
) -> EllipticCurvePublicKey:
    """Generate ECC private key.

    :param private_key:
    :return: ECC public key
    """
    return private_key.public_key()


def save_ecc_private_key(
    ec_private_key: EllipticCurvePrivateKeyWithSerialization,
    file_path: str,
    password: Optional[str] = None,
    encoding: Encoding = Encoding.PEM,
) -> None:
    """Save the ECC private key to the given file.

    :param ec_private_key: ECC private key to be saved
    :param file_path: path to the file, where the key will be stored
    :param password: password to private key; None to store without password
    :param encoding: encoding type, default is PEM
    """
    serialized_private = ec_private_key.private_bytes(
        encoding=encoding,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.BestAvailableEncryption(password.encode("utf-8"))
        if password
        else serialization.NoEncryption(),
    )
    write_file(serialized_private, file_path, mode="wb")


def save_ecc_public_key(
    ec_public_key: EllipticCurvePublicKey,
    file_path: str,
    encoding: Encoding = Encoding.PEM,
) -> None:
    """Save the ECC public key to the file.

    :param ec_public_key: public key to be saved
    :param file_path: path to the file, where the key will be stored
    :param encoding: encoding type, default is PEM
    """
    pem_data = ec_public_key.public_bytes(
        encoding=encoding, format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    write_file(pem_data, file_path, mode="wb")


def recreate_ecc_public_key(coor_x: int, coor_y: int, curve: str) -> EllipticCurvePublicKey:
    """Recreate ECC public key from coordinates.

    :param coor_x: X coordinate of point on curve.
    :param coor_y: Y coordinate of point on curve.
    :param curve: Name of ECC curve.
    :return: ECC public key.
    """
    pub_numbers = ec.EllipticCurvePublicNumbers(
        x=coor_x, y=coor_y, curve=get_ec_curve_object(curve)
    )
    key = pub_numbers.public_key()
    return key


def recreate_ecc_public_key_from_data(coors: bytes, curve: str) -> EllipticCurvePublicKey:
    """Recreate ECC public key from coordinates in data blob.

    :param coors: Data blob of coordinates in bytes (X,Y in Big Endian)
    :param curve: Name of ECC curve.
    :return: ECC public key.
    """
    coordinate_length = len(coors) // 2
    coor_x = int.from_bytes(coors[:coordinate_length], byteorder="big")
    coor_y = int.from_bytes(coors[coordinate_length:], byteorder="big")
    return recreate_ecc_public_key(coor_x=coor_x, coor_y=coor_y, curve=curve)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for handling Certificate block."""

import datetime
import logging
import os
import re
from struct import calcsize, pack, unpack_from
from typing import Any, Dict, List, Optional, Sequence, Union

from Crypto.PublicKey import ECC

from spsdk import SPSDKError
from spsdk.crypto import loaders
from spsdk.crypto.loaders import load_certificate_as_bytes
from spsdk.crypto.signature_provider import SignatureProvider, get_signature_provider
from spsdk.exceptions import SPSDKValueError
from spsdk.utils import misc
from spsdk.utils.crypto import CRYPTO_SCH_FILE
from spsdk.utils.schema_validator import CommentedConfig, ValidationSchemas

from .abstract import BaseClass
from .backend_internal import internal_backend
from .certificate import Certificate
from .common import crypto_backend, get_matching_key_id

logger = logging.getLogger(__name__)


class CertBlock(BaseClass):
    """Common general class for various CertBlocks."""


########################################################################################################################
# Certificate Block Header Class
########################################################################################################################
class CertBlockHeader(BaseClass):
    """Certificate block header."""

    FORMAT = "<4s2H6I"
    SIZE = calcsize(FORMAT)
    SIGNATURE = b"cert"

    def __init__(self, version: str = "1.0", flags: int = 0, build_number: int = 0) -> None:
        """Constructor.

        :param version: Version of the certificate in format n.n
        :param flags: Flags for the Certificate Header
        :param build_number: of the certificate
        :raises SPSDKError: When there is invalid version
        """
        if not re.match(r"[0-9]+\.[0-9]+", version):  # check format of the version: N.N
            raise SPSDKError("Invalid version")
        self.version = version
        self.flags = flags
        self.build_number = build_number
        self.image_length = 0
        self.cert_count = 0
        self.cert_table_length = 0

    def __str__(self) -> str:
        nfo = f"CertBlockHeader: V={self.version}, F={self.flags}, BN={self.build_number}, IL={self.image_length}, "
        nfo += f"CC={self.cert_count}, CTL={self.cert_table_length}"
        return nfo

    def info(self) -> str:
        """Info of the certificate header in text form."""
        nfo = str()
        nfo += f" CB Version:           {self.version}\n"
        nfo += f" CB Flags:             {self.flags}\n"
        nfo += f" CB Build Number:      {self.build_number}\n"
        nfo += f" CB Image Length:      {self.image_length}\n"
        nfo += f" CB Cert. Count:       {self.cert_count}\n"
        nfo += f" CB Cert. Length:      {self.cert_table_length}\n"
        return nfo

    def export(self) -> bytes:
        """Certificate block in binary form."""
        major_version, minor_version = [int(v) for v in self.version.split(".")]
        return pack(
            self.FORMAT,
            self.SIGNATURE,
            major_version,
            minor_version,
            self.SIZE,
            self.flags,
            self.build_number,
            self.image_length,
            self.cert_count,
            self.cert_table_length,
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CertBlockHeader":
        """Deserialize object from bytes array.

        :param data: Input data as bytes
        :param offset: The offset of input data (default: 0)
        :return: Certificate Header instance
        :raises SPSDKError: Unexpected size or signature of data
        """
        if cls.SIZE > len(data) - offset:
            raise SPSDKError("Incorrect size")
        (
            signature,
            major_version,
            minor_version,
            length,
            flags,
            build_number,
            image_length,
            cert_count,
            cert_table_length,
        ) = unpack_from(cls.FORMAT, data, offset)
        if signature != cls.SIGNATURE:
            raise SPSDKError("Incorrect signature")
        if length != cls.SIZE:
            raise SPSDKError("Incorrect length")
        obj = cls(
            version=f"{major_version}.{minor_version}",
            flags=flags,
            build_number=build_number,
        )
        obj.image_length = image_length
        obj.cert_count = cert_count
        obj.cert_table_length = cert_table_length
        return obj


########################################################################################################################
# Certificate Block Class
########################################################################################################################
class CertBlockV2(CertBlock):
    """Certificate block.

    Shared for SB file and for MasterBootImage
    """

    # size of the hash in bytes (single item in RKH table)
    RKH_SIZE = 32
    # number of hashes in RKH table (number of table entries)
    RKHT_SIZE = 4

    # default size alignment
    DEFAULT_ALIGNMENT = 16

    @property
    def header(self) -> CertBlockHeader:
        """Certificate block header."""
        return self._header

    @property
    def rkh(self) -> List[bytes]:
        """List of root keys hashes (SHA-256), each hash as 32 bytes."""
        return self._root_key_hashes

    @property
    def rkht(self) -> bytes:
        """32-byte hash (SHA-256) of SHA-256 hashes of up to four root public keys."""
        data = bytes()
        for rkh in self._root_key_hashes:
            data += rkh
        if len(data) != self.RKH_SIZE * self.RKHT_SIZE:
            raise SPSDKError("Invalid length of data")
        return internal_backend.hash(data)

    @property
    def rkht_fuses(self) -> List[int]:
        """List of RKHT fuses, ordered from highest bit to lowest.

        Note: Returned values are in format that should be passed for blhost
        """
        result = []
        rkht = self.rkht
        while rkht:
            fuse = int.from_bytes(rkht[:4], byteorder="little")
            result.append(fuse)
            rkht = rkht[4:]
        return result

    @property
    def certificates(self) -> List[Certificate]:
        """List of certificates in header.

        First certificate is root certificate and followed by optional chain certificates
        """
        return self._cert

    @property
    def signature_size(self) -> int:
        """Size of the signature in bytes."""
        return len(
            self.certificates[0].signature
        )  # The certificate is self signed, return size of its signature

    @property
    def rkh_index(self) -> Optional[int]:
        """Index of the hash that matches the certificate; None if does not match."""
        if self._cert:
            rkh = self._cert[0].public_key_hash
            for index, value in enumerate(self._root_key_hashes):
                if rkh == value:
                    return index
        return None

    @property
    def alignment(self) -> int:
        """Alignment of the binary output, by default it is DEFAULT_ALIGNMENT but can be customized."""
        return self._alignment

    @alignment.setter
    def alignment(self, value: int) -> None:
        """Setter.

        :param value: new alignment
        :raises SPSDKError: When there is invalid alignment
        """
        if value <= 0:
            raise SPSDKError("Invalid alignment")
        self._alignment = value

    @property
    def raw_size(self) -> int:
        """Aligned size of the certificate block."""
        size = CertBlockHeader.SIZE
        size += self._header.cert_table_length
        size += self.RKH_SIZE * self.RKHT_SIZE
        return misc.align(size, self.alignment)

    @property
    def expected_size(self) -> int:
        """Expected size of binary block."""
        return self.raw_size

    @property
    def image_length(self) -> int:
        """Image length in bytes."""
        return self._header.image_length

    @image_length.setter
    def image_length(self, value: int) -> None:
        """Setter.

        :param value: new image length
        :raises SPSDKError: When there is invalid image length
        """
        if value <= 0:
            raise SPSDKError("Invalid image length")
        self._header.image_length = value

    def __init__(self, version: str = "1.0", flags: int = 0, build_number: int = 0) -> None:
        """Constructor.

        :param version: of the certificate in format n.n
        :param flags: Flags for the Certificate Block Header
        :param build_number: of the certificate
        """
        self._header = CertBlockHeader(version, flags, build_number)
        self._root_key_hashes = [b"\x00" * self.RKH_SIZE for _ in range(self.RKHT_SIZE)]
        self._cert: List[Certificate] = []
        self._alignment = self.DEFAULT_ALIGNMENT

    def __str__(self) -> str:
        return str(self._header)

    def __len__(self) -> int:
        return len(self._cert)

    def set_root_key_hash(self, index: int, key_hash: Union[bytes, bytearray, Certificate]) -> None:
        """Add Root Key Hash into RKH at specified index.

        Note: Multiple root public keys are supported to allow for key revocation.

        :param index: The index of Root Key Hash in the table
        :param key_hash: The Root Key Hash value (32 bytes, SHA-256);
                        or Certificate where the hash can be created from public key
        :raises SPSDKError: When there is invalid index of root key hash in the table
        :raises SPSDKError: When there is invalid length of key hash
        """
        if isinstance(key_hash, Certificate):
            key_hash = key_hash.public_key_hash
        assert isinstance(key_hash, (bytes, bytearray))
        if index < 0 or index >= self.RKHT_SIZE:
            raise SPSDKError("Invalid index of root key hash in the table")
        if len(key_hash) != self.RKH_SIZE:
            raise SPSDKError("Invalid length of key hash")
        self._root_key_hashes[index] = bytes(key_hash)

    def add_certificate(self, cert: Union[bytes, Certificate]) -> None:
        """Add certificate.

        First call adds root certificate. Additional calls add chain certificates.

        :param cert: The certificate itself in DER format
        :raises SPSDKError: If certificate cannot be added
        """
        if isinstance(cert, bytes):
            cert_obj = Certificate(cert)
        elif isinstance(cert, Certificate):
            cert_obj = cert
        else:
            raise SPSDKError("Invalid parameter type (cert)")
        if cert_obj.version != "v3":
            raise SPSDKError("Expected certificate v3 but received: " + cert_obj.version)
        if self._cert:  # chain certificate?
            last_cert = self._cert[-1]  # verify that it is signed by parent key
            if not cert_obj.verify(last_cert.public_key_modulus, last_cert.public_key_exponent):
                raise SPSDKError("Chain certificate cannot be verified using parent public key")
        else:  # root certificate
            if cert_obj.self_signed == "no":
                raise SPSDKError(f"Root certificate must be self-signed.\n{cert_obj.info()}")
        self._cert.append(cert_obj)
        self._header.cert_count += 1
        self._header.cert_table_length += cert_obj.raw_size + 4

    def info(self) -> str:
        """Text info about certificate block."""
        nfo = self.header.info()
        nfo += " Public Root Keys Hash e.g. RKH (SHA256):\n"
        rkh_index = self.rkh_index
        for index, root_key in enumerate(self._root_key_hashes):
            nfo += (
                f"  {index}) {root_key.hex().upper()} {'<- Used' if index == rkh_index else ''}\n"
            )
        rkht = self.rkht
        nfo += f" RKHT (SHA256): {rkht.hex().upper()}\n"
        for index, fuse in enumerate(self.rkht_fuses):
            bit_ofs = (len(rkht) - 4 * index) * 8
            nfo += f"  - RKHT fuse [{bit_ofs:03}:{bit_ofs - 31:03}]: {fuse:08X}\n"
        for index, cert in enumerate(self._cert):
            nfo += " Root Certificate:\n" if index == 0 else f" Certificate {index}:\n"
            nfo += cert.info()
        return nfo

    def verify_data(self, signature: bytes, data: bytes) -> bool:
        """Signature verification.

        :param signature: to be verified
        :param data: that has been signed
        :return: True if the data signature can be confirmed using the certificate; False otherwise
        """
        cert = self._cert[-1]
        return crypto_backend().rsa_verify(
            cert.public_key_modulus, cert.public_key_exponent, signature, data
        )

    def verify_private_key(self, private_key_pem_data: bytes) -> bool:
        """Verify that given private key matches the public certificate.

        :param private_key_pem_data: to be tested; decrypted binary data in PEM format
        :return: True if yes; False otherwise
        """
        signature = crypto_backend().rsa_sign(private_key_pem_data, bytes())
        cert = self.certificates[-1]  # last certificate
        return crypto_backend().rsa_verify(
            cert.public_key_modulus, cert.public_key_exponent, signature, bytes()
        )

    def export(self) -> bytes:
        """Serialize Certificate Block V2 object."""
        # At least one certificate must be used
        if not self._cert:
            raise SPSDKError("At least one certificate must be used")
        # The hast of root key certificate must be in RKHT
        if self.rkh_index is None:
            raise SPSDKError("The HASH of used Root Key must be in RKHT")
        # CA: Using a single certificate is allowed. In this case, the sole certificate must be self-signed and must not
        # be a CA. If multiple certificates are used, the root must be self-signed and all but the last must be CAs.
        if self._cert[-1].ca:
            raise SPSDKError("The last chain certificate must not be CA.")
        if not all(cert.ca for cert in self._cert[:-1]):
            raise SPSDKError("All certificates except the last chain certificate must be CA")
        # Export
        data = self.header.export()
        for cert in self._cert:
            data += pack("<I", cert.raw_size)
            data += cert.export()
        for key in self._root_key_hashes:
            data += bytes(key)
        data = misc.align_block(data, self.alignment)
        if len(data) != self.raw_size:
            raise SPSDKError("Invalid length of data")
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CertBlockV2":
        """Deserialize CertBlockV2 from binary file.

        :param data: Binary data
        :param offset: Offset within the data, where the Certificate block begins, defaults to 0
        :return: Certificate Block instance
        :raises SPSDKError: Length of the data doesn't match Certificate Block length
        """
        header = CertBlockHeader.parse(data, offset)
        offset += CertBlockHeader.SIZE
        if (len(data) - offset) < (header.cert_table_length + (cls.RKHT_SIZE * cls.RKH_SIZE)):
            raise SPSDKError("Length of the data doesn't match Certificate Block length")
        obj = cls(version=header.version, flags=header.flags, build_number=header.build_number)
        for i in range(header.cert_count):
            cert_len = unpack_from("<I", data, offset)[0]
            offset += 4
            cert_obj = Certificate(data[offset : offset + cert_len])
            obj.add_certificate(cert_obj)
            offset += cert_len
        for i in range(cls.RKHT_SIZE):
            obj.set_root_key_hash(i, data[offset : offset + cls.RKH_SIZE])
            offset += cls.RKH_SIZE
        return obj

    @classmethod
    def get_validation_schemas(cls) -> List[Dict[str, Any]]:
        """Create the list of validation schemas.

        :return: List of validation schemas.
        """
        sch_cfg = ValidationSchemas.get_schema_file(CRYPTO_SCH_FILE)
        return [
            sch_cfg["certificate_v2"],
            sch_cfg["certificate_root_keys"],
        ]

    @classmethod
    def from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CertBlockV2":
        """Creates an instance of CertBlockV2 from configuration.

        :param config: Input standard configuration.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Instance of CertBlockV2
        :raises SPSDKError: Invalid certificates detected.
        """
        image_build_number = misc.value_to_int(config.get("imageBuildNumber", 0))
        root_certificates: List[List[str]] = [[] for _ in range(4)]
        # TODO we need to read the whole chain from the dict for a given
        # selection based on mainCertPrivateKeyFile!!!
        root_certificates[0].append(config.get("rootCertificate0File", None))
        root_certificates[1].append(config.get("rootCertificate1File", None))
        root_certificates[2].append(config.get("rootCertificate2File", None))
        root_certificates[3].append(config.get("rootCertificate3File", None))
        main_cert_chain_id = get_main_cert_index(config, search_paths=search_paths)
        if root_certificates[main_cert_chain_id][0] is None:
            raise SPSDKError(f"A key rootCertificate{main_cert_chain_id}File must be defined")

        # get all certificate chain related keys from config
        pattern = f"chainCertificate{main_cert_chain_id}File[0-3]"
        keys = [key for key in config.keys() if re.fullmatch(pattern, key)]
        # just in case, sort the chain certificate keys in order
        keys.sort()
        for key in keys:
            root_certificates[main_cert_chain_id].append(config[key])

        cert_block = CertBlockV2(build_number=image_build_number)

        # add whole certificate chain used for image signing
        for cert_path in root_certificates[main_cert_chain_id]:
            cert_data = load_certificate_as_bytes(
                misc.find_file(str(cert_path), search_paths=search_paths)
            )
            cert_block.add_certificate(cert_data)
        # set root key hash of each root certificate
        empty_rec = False
        for cert_idx, cert_path_list in enumerate(root_certificates):
            if cert_path_list[0]:
                if empty_rec:
                    raise SPSDKError("There are gaps in rootCertificateXFile definition")
                cert_data = load_certificate_as_bytes(
                    misc.find_file(str(cert_path_list[0]), search_paths=search_paths)
                )
                cert_block.set_root_key_hash(cert_idx, Certificate(cert_data))
            else:
                empty_rec = True

        return cert_block

    def get_config(self, output_folder: str) -> Dict[str, Any]:
        """Create configuration of Certificate V2 from object.

        :param output_folder: Output folder to store possible files.
        :return: Configuration dictionary.
        """

        def create_certificate_cfg(root_id: int, chain_id: int) -> Optional[str]:
            if len(self._cert) <= chain_id:
                return None

            file_name = f"certificate{root_id}_depth{chain_id}.der"
            misc.write_file(
                self._cert[chain_id].dump(), os.path.join(output_folder, file_name), mode="wb"
            )
            return file_name

        cfg: Dict[str, Optional[Union[str, int]]] = {}
        cfg["imageBuildNumber"] = self.header.build_number
        used_cert_id = self.rkh_index
        assert used_cert_id is not None
        cfg["mainRootCertId"] = used_cert_id

        cfg[f"rootCertificate{used_cert_id}File"] = create_certificate_cfg(used_cert_id, 0)
        for chain_ix in range(4):
            cfg[f"chainCertificate{used_cert_id}File{chain_ix}"] = create_certificate_cfg(
                used_cert_id, chain_ix + 1
            )

        return cfg


########################################################################################################################
# Certificate Block Class for SB 3.1
########################################################################################################################
def get_ecc_key_bytes(key: ECC.EccKey) -> bytes:
    """Function to get ECC Key pointQ as bytes."""
    point_x = key.pointQ.x.to_bytes(block_size=key.pointQ.size_in_bytes())  # type: ignore
    point_y = key.pointQ.y.to_bytes(block_size=key.pointQ.size_in_bytes())  # type: ignore
    return point_x + point_y


def convert_to_ecc_key(key: Union[ECC.EccKey, bytes]) -> ECC.EccKey:
    """Convert key into EccKey instance."""
    if isinstance(key, ECC.EccKey):
        return key
    try:
        return ECC.import_key(key)
    except Exception:
        pass
    # Just recreate public key from the parsed data
    coordinate_length = len(key) // 2
    coor_x = int.from_bytes(key[:coordinate_length], byteorder="big")
    coor_y = int.from_bytes(key[coordinate_length:], byteorder="big")
    curve = "secp256r1" if coordinate_length == 32 else "secp384r1"
    ecc_point = ECC.EccPoint(coor_x, coor_y, curve)
    return ECC.EccKey(curve=curve, point=ecc_point)


class CertificateBlockHeader(BaseClass):
    """Create Certificate block header."""

    FORMAT = "<4s2HL"
    SIZE = calcsize(FORMAT)
    MAGIC = b"chdr"

    def __init__(self, format_version: str = "2.1") -> None:
        """Constructor for Certificate block header version 2.1.

        :param format_version: Major = 2, minor = 1
        """
        self.format_version = format_version
        self.cert_block_size = 0

    def info(self) -> str:
        """Get info of Certificate block header."""
        info = f"Format version:              {self.format_version}\n"
        info += f"Certificate block size:      {self.cert_block_size}\n"
        return info

    def export(self) -> bytes:
        """Export Certificate block header as bytes array."""
        major_format_version, minor_format_version = [
            int(v) for v in self.format_version.split(".")
        ]

        return pack(
            self.FORMAT,
            self.MAGIC,
            minor_format_version,
            major_format_version,
            self.cert_block_size,
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CertificateBlockHeader":
        """Parse Certificate block header from bytes array.

        :param data: Input data as bytes
        :param offset: The offset of input data (default: 0)
        :raises SPSDKError: Raised when SIZE is bigger than length of the data without offset
        :raises SPSDKError: Raised when magic is not equal MAGIC
        :return: CertificateBlockHeader
        """
        if cls.SIZE > len(data) - offset:
            raise SPSDKError("SIZE is bigger than length of the data without offset")
        (
            magic,
            minor_format_version,
            major_format_version,
            cert_block_size,
        ) = unpack_from(cls.FORMAT, data, offset)

        if magic != cls.MAGIC:
            raise SPSDKError("Magic is not same!")

        obj = cls(format_version=f"{major_format_version}.{minor_format_version}")
        obj.cert_block_size = cert_block_size
        return obj

    def __len__(self) -> int:
        """Length of the Certificate block header."""
        return calcsize(self.FORMAT)


class RootKeyRecord(BaseClass):
    """Create Root key record."""

    # P-256

    def __init__(
        self,
        ca_flag: bool,
        root_certs: Optional[Union[Sequence[ECC.EccKey], Sequence[bytes]]] = None,
        used_root_cert: int = 0,
    ) -> None:
        """Constructor for Root key record.

        :param ca_flag: CA flag
        :param root_certs: Root cert used to ISK/image signature
        :param used_root_cert: Used root cert number 0-3
        """
        self.ca_flag = ca_flag
        self.root_certs_input = root_certs
        self.root_certs: List[ECC.EccKey] = []
        self.used_root_cert = used_root_cert
        self.flags = 0
        self.ctrk_hash_table = b""
        self.rotkth = b""
        self.root_public_key = b""

    @property
    def number_of_certificates(self) -> int:
        """Get number of included certificates."""
        return (self.flags & 0xF0) >> 4

    @property
    def expected_size(self) -> int:
        """Get expected binary block size."""
        # the '4' means 4 bytes for flags
        return 4 + len(self.ctrk_hash_table) + len(self.root_public_key)

    def info(self) -> str:
        """Get info of Root key record."""
        cert_type = {0x1: "NIST P-256", 0x2: "NIST P-384"}[self.flags & 0xF]
        info = ""
        info += f"Flags:           {hex(self.flags)}\n"
        info += f"  - CA:          {bool(self.ca_flag)}, ISK Certificate is {'not ' if self.ca_flag else ''}mandatory\n"
        info += f"  - Used Root c.:{self.used_root_cert}\n"
        info += f"  - Number of c.:{self.number_of_certificates}\n"
        info += f"  - Cert. type:  {cert_type}\n"
        if self.root_certs:
            info += f"Root certs:      {self.root_certs}\n"
        if self.ctrk_hash_table:
            info += f"CTRK Hash table: {self.ctrk_hash_table.hex()}\n"
        if self.root_public_key:
            info += f"Root public key: {str(convert_to_ecc_key(self.root_public_key))}\n"

        return info

    def _calculate_flags(self) -> int:
        """Function to calculate parameter flags."""
        flags = 0
        if self.ca_flag is True:
            flags |= 1 << 31
        if self.used_root_cert:
            flags |= self.used_root_cert << 8
        flags |= len(self.root_certs) << 4
        if self.root_certs[0].curve in ["NIST P-256", "p256", "secp256r1"]:
            flags |= 1 << 0
        if self.root_certs[0].curve in ["NIST P-384", "p384", "secp384r1"]:
            flags |= 1 << 1
        return flags

    def _create_root_public_key(self) -> bytes:
        """Function to create root public key."""
        root_key = self.root_certs[self.used_root_cert]
        root_key_data = get_ecc_key_bytes(root_key)
        return root_key_data

    def _create_ctrk_hash_table(self) -> bytes:
        """Function to create ctrk hash table."""
        ctrk_hash_table = bytes()
        if len(self.root_certs) > 1:
            for key in self.root_certs:
                data_to_hash = get_ecc_key_bytes(key)
                ctrk_hash = internal_backend.hash(
                    data=data_to_hash, algorithm=self.get_hash_algorithm(self.flags)
                )
                ctrk_hash_table += ctrk_hash
        return ctrk_hash_table

    def _calculate_rotkth(self) -> bytes:
        return internal_backend.hash(
            data=self.ctrk_hash_table, algorithm=self.get_hash_algorithm(self.flags)
        )

    def calculate(self) -> None:
        """Calculate all internal members.

        :raises SPSDKError: The RoT certificates inputs are missing.
        """
        # pylint: disable=invalid-name
        if not self.root_certs_input:
            raise SPSDKError("Root Key Record: The root of trust certificates are not specified.")
        self.root_certs = [convert_to_ecc_key(cert) for cert in self.root_certs_input]
        self.flags = self._calculate_flags()
        self.ctrk_hash_table = self._create_ctrk_hash_table()
        self.rotkth = self._calculate_rotkth()
        self.root_public_key = self._create_root_public_key()

    def export(self) -> bytes:
        """Export Root key record as bytes array."""
        data = bytes()
        data += pack("<L", self.flags)
        data += self.ctrk_hash_table
        data += self.root_public_key
        assert len(data) == self.expected_size
        return data

    @staticmethod
    def get_hash_algorithm(flags: int) -> str:
        """Get CTRK table hash algorithm.

        :param flags: Root Key Record flags
        :return: Name of hash algorithm
        """
        return {1: "sha256", 2: "sha384"}[flags & 0xF]

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "RootKeyRecord":
        """Parse Root key record from bytes array.

        :param data:  Input data as bytes array
        :param offset: The offset of input data
        :return: Root key record object
        """
        (flags,) = unpack_from("<L", data, offset=offset)
        ca_flag = flags & 0x80000000
        used_rot_ix = (flags & 0xF00) >> 8
        number_of_hashes = (flags & 0xF0) >> 4
        rotkh_len = {0x1: 32, 0x2: 48}[flags & 0xF]
        root_key_record = RootKeyRecord(ca_flag=ca_flag, root_certs=[], used_root_cert=used_rot_ix)
        root_key_record.flags = flags
        offset += 4  # move offset just after FLAGS
        if number_of_hashes > 1:
            root_key_record.ctrk_hash_table = data[offset : offset + rotkh_len * number_of_hashes]
            offset += rotkh_len * number_of_hashes
        root_key_record._calculate_rotkth()
        root_key_record.root_public_key = data[offset : offset + rotkh_len * 2]

        return root_key_record


class IskCertificate(BaseClass):
    """Create ISK certificate."""

    def __init__(
        self,
        constraints: int = 0,
        signature_provider: Optional[SignatureProvider] = None,
        isk_cert: Optional[Union[ECC.EccKey, bytes]] = None,
        user_data: Optional[bytes] = None,
    ) -> None:
        """Constructor for ISK certificate.

        :param constraints: Certificate version
        :param signature_provider: ISK Signature Provider
        :param isk_cert: ISK certificate
        :param user_data: User data
        """
        self.flags = 0
        self.constraints = constraints
        self.signature_provider = signature_provider
        self.isk_cert = convert_to_ecc_key(isk_cert) if isk_cert else None
        self.user_data = user_data or bytes()
        self.signature = bytes()
        self.coordinate_length = (
            self.signature_provider.signature_length // 2 if self.signature_provider else 0
        )
        self.isk_public_key_data = get_ecc_key_bytes(self.isk_cert) if self.isk_cert else bytes()

        self._calculate_flags()

    @property
    def signature_offset(self) -> int:
        """Signature offset inside the ISK Certificate."""
        signature_offset = calcsize("<3L") + len(self.user_data)
        if self.isk_cert:
            signature_offset += 2 * self.isk_cert.pointQ.size_in_bytes()

        return signature_offset

    @property
    def expected_size(self) -> int:
        """Binary block expected size."""
        sign_len = len(self.signature) or (
            self.signature_provider.signature_length if self.signature_provider else 0
        )
        pub_key_len = (
            self.isk_cert.pointQ.size_in_bytes() * 2
            if self.isk_cert
            else len(self.isk_public_key_data)
        )

        return (
            4  #  signature offset
            + 4  # constraints
            + 4  # flags
            + pub_key_len  # isk public key coordinates
            + len(self.user_data)  # user data
            + sign_len  # isk blob signature
        )

    def info(self) -> str:
        """Get info of ISK certificate."""
        isk_type = {0x1: "NIST P-256", 0x2: "NIST P-384"}[self.flags & 0xF]
        info = ""
        info += f"Constraints:     {self.constraints}\n"
        if self.user_data:
            info += f"User data:       {self.user_data.hex()}\n"
        else:
            info += "User data:       Not included\n"
        info += f"Type:            {isk_type}\n"
        info += f"Public Key:      {str(self.isk_cert)}\n"
        return info

    def _calculate_flags(self) -> None:
        """Function to calculate parameter flags."""
        self.flags = 0
        if self.user_data:
            self.flags |= 1 << 31
        assert self.isk_cert
        if self.isk_cert.curve in ["NIST P-256", "p256", "secp256r1"]:
            self.flags |= 1 << 0
        if self.isk_cert.curve in ["NIST P-384", "p384", "secp384r1"]:
            self.flags |= 1 << 1

    def create_isk_signature(self, key_record_data: bytes, force: bool = False) -> None:
        """Function to create ISK signature.

        :raises SPSDKError: Signature provider is not specified.
        """
        # pylint: disable=invalid-name
        if self.signature and not force:
            return
        if not self.signature_provider:
            raise SPSDKError("ISK Certificate: The signature provider is not specified.")
        data = key_record_data + pack("<3L", self.signature_offset, self.constraints, self.flags)
        data += self.isk_public_key_data + self.user_data
        self.signature = self.signature_provider.sign(data)

    def export(self) -> bytes:
        """Export ISK certificate as bytes array."""
        if not self.signature:
            raise SPSDKError("Signature is not set.")
        data = pack("<3L", self.signature_offset, self.constraints, self.flags)
        data += self.isk_public_key_data
        if self.user_data:
            data += self.user_data
        data += self.signature

        assert len(data) == self.expected_size
        return data

    @classmethod
    def parse(  # type: ignore
        cls, data: bytes, signature_size: int, offset: int = 0
    ) -> "IskCertificate":
        """Parse ISK certificate from bytes array.This operation is not supported.

        :param data:  Input data as bytes array
        :param signature_size: The signature size of ISK block
        :param offset: The offset of input data
        :raises NotImplementedError: This operation is not supported
        """
        (signature_offset, constraints, isk_flags) = unpack_from("<3L", data, offset)
        signature_offset += offset
        user_data_flag = bool(isk_flags & 0x80000000)
        isk_pub_key_length = {0x1: 32, 0x2: 48}[isk_flags & 0xF]
        offset += 3 * 4
        isk_pub_key_bytes = data[offset : offset + isk_pub_key_length * 2]
        offset += isk_pub_key_length * 2
        user_data = data[offset:signature_offset] if user_data_flag else None
        signature = data[signature_offset : signature_offset + signature_size]

        certificate = IskCertificate(
            constraints=constraints, isk_cert=isk_pub_key_bytes, user_data=user_data
        )
        certificate.signature = signature
        return certificate


class CertBlockV31(CertBlock):
    """Create Certificate block version 3.1."""

    MAGIC = b"chdr"
    FORMAT_VERSION = "2.1"

    def __init__(
        self,
        root_certs: Optional[Union[Sequence[ECC.EccKey], Sequence[bytes]]] = None,
        ca_flag: bool = False,
        version: str = "2.1",
        used_root_cert: int = 0,
        constraints: int = 0,
        signature_provider: Optional[SignatureProvider] = None,
        isk_cert: Optional[Union[ECC.EccKey, bytes]] = None,
        user_data: Optional[bytes] = None,
    ) -> None:
        """The Constructor for Certificate block."""
        self.header = CertificateBlockHeader(version)
        self.root_key_record = RootKeyRecord(
            ca_flag=ca_flag, used_root_cert=used_root_cert, root_certs=root_certs
        )

        self.isk_certificate = None
        if not ca_flag and signature_provider and isk_cert:
            self.isk_certificate = IskCertificate(
                constraints=constraints,
                signature_provider=signature_provider,
                isk_cert=isk_cert,
                user_data=user_data,
            )

    def _set_ca_flag(self, value: bool) -> None:
        self.root_key_record.ca_flag = value

    def calculate(self) -> None:
        """Calculate all internal members."""
        self.root_key_record.calculate()

    @property
    def signature_size(self) -> int:
        """Size of the signature in bytes."""
        # signature size is same as public key data
        if self.isk_certificate:
            return len(self.isk_certificate.isk_public_key_data)

        return len(self.root_key_record.root_public_key)

    @property
    def expected_size(self) -> int:
        """Expected size of binary block."""
        expected_size = self.header.SIZE
        expected_size += self.root_key_record.expected_size
        if self.isk_certificate:
            expected_size += self.isk_certificate.expected_size
        return expected_size

    @property
    def rkht(self) -> bytes:
        """32-byte hash (SHA-256) of SHA-256 hashes of up to four root public keys."""
        return self.root_key_record.rotkth

    def info(self) -> str:
        """Get info of Certificate block."""
        msg = f"HEADER:\n{self.header.info()}\n"
        msg += f"ROOT KEY RECORD:\n{self.root_key_record.info()}\n"
        if self.isk_certificate:
            msg += f"ISK Certificate:\n{self.isk_certificate.info()}\n"
        return msg

    def export(self) -> bytes:
        """Export Certificate block as bytes array."""
        logger.info(f"RoTKTH: {self.root_key_record.rotkth.hex()}")
        key_record_data = self.root_key_record.export()
        self.header.cert_block_size = self.header.SIZE + len(key_record_data)
        isk_cert_data = bytes()
        if self.isk_certificate:
            self.isk_certificate.create_isk_signature(key_record_data)
            isk_cert_data = self.isk_certificate.export()
            self.header.cert_block_size += len(isk_cert_data)
        header_data = self.header.export()
        return header_data + key_record_data + isk_cert_data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CertBlockV31":
        """Parse Certificate block from bytes array.This operation is not supported.

        :param data:  Input data as bytes array
        :param offset: The offset of input data
        :raises SPSDKError: Magic do not match
        """
        # CertificateBlockHeader
        cert_header = CertificateBlockHeader.parse(data, offset)
        offset += len(cert_header)
        # RootKeyRecord
        root_key_record = RootKeyRecord.parse(data, offset)
        offset += root_key_record.expected_size
        # IskCertificate
        isk_certificate = None
        if root_key_record.ca_flag == 0:
            isk_certificate = IskCertificate.parse(
                data, len(root_key_record.root_public_key), offset
            )
        # Certification Block V3.1
        cert_block = CertBlockV31()
        cert_block.header = cert_header
        cert_block.root_key_record = root_key_record
        cert_block.isk_certificate = isk_certificate
        return cert_block

    @classmethod
    def get_validation_schemas(cls) -> List[Dict[str, Any]]:
        """Create the list of validation schemas.

        :return: List of validation schemas.
        """
        sch_cfg = ValidationSchemas.get_schema_file(CRYPTO_SCH_FILE)
        return [sch_cfg["certificate_v31"], sch_cfg["certificate_root_keys"]]

    @classmethod
    def from_config(
        cls, config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "CertBlockV31":
        """Creates an instance of CertBlockV31 from configuration.

        :param config: Input standard configuration.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Instance of CertBlockV3.1
        :raises SPSDKError: If found gap in certificates from config file.
        """
        binary_block = config.get("binaryCertificateBlock")
        if binary_block:
            return CertBlockV31.parse(misc.load_binary(binary_block, search_paths))

        root_certificates = find_root_certificates(config)
        main_root_cert_id = get_main_cert_index(config, search_paths=search_paths)

        try:
            root_certificates[main_root_cert_id]
        except IndexError as e:
            raise SPSDKError(
                f"Main root certificate with id {main_root_cert_id} does not exist"
            ) from e

        main_root_private_key_file = config.get("mainRootCertPrivateKeyFile")
        signature_provider = config.get("iskSignProvider")
        use_isk = config.get("useIsk", False)
        isk_certificate = config.get("signingCertificateFile")
        isk_constraint = misc.value_to_int(config.get("signingCertificateConstraint", "0"))
        isk_sign_data_path = config.get("signCertData")

        root_certs = [
            misc.load_binary(cert_file, search_paths=search_paths)
            for cert_file in root_certificates
        ]
        user_data = None
        signature_provider = None
        isk_cert = None

        if use_isk:
            assert isk_certificate and (main_root_private_key_file or signature_provider)
            if isk_sign_data_path:
                user_data = misc.load_binary(isk_sign_data_path, search_paths=search_paths)
            signature_provider = get_signature_provider(
                signature_provider,
                main_root_private_key_file,
                search_paths=search_paths,
                mode="deterministic-rfc6979",
            )
            isk_cert = misc.load_binary(isk_certificate, search_paths=search_paths)

        cert_block = CertBlockV31(
            root_certs=root_certs,
            used_root_cert=main_root_cert_id,
            user_data=user_data,
            constraints=isk_constraint,
            isk_cert=isk_cert,
            ca_flag=not use_isk,
            signature_provider=signature_provider,
        )
        cert_block.calculate()

        return cert_block

    def validate(self) -> None:
        """Validate the settings of class members.

        :raises SPSDKError: Invalid configuration of certification block class members.
        """
        self.header.parse(self.header.export())
        if self.isk_certificate:
            if not isinstance(self.isk_certificate.signature_provider, SignatureProvider):
                raise SPSDKError("Invalid ISK certificate.")

    @staticmethod
    def generate_config_template() -> str:
        """Generate configuration for certification block v31."""
        val_schemas = CertBlockV31.get_validation_schemas()
        val_schemas.append(ValidationSchemas.get_schema_file(CRYPTO_SCH_FILE)["cert_block_output"])
        yaml_data = CommentedConfig(
            "Certification Block V31 template",
            val_schemas,
        ).export_to_yaml()
        return yaml_data

    def get_config(self, output_folder: str) -> Dict[str, Any]:
        """Create configuration dictionary of the Certification block Image.

        :param output_folder: Path to store the data files of configuration.
        :return: Configuration dictionary.
        """
        cfg: Dict[str, Optional[Union[str, int]]] = {}
        cfg["mainRootCertPrivateKeyFile"] = "N/A"
        cfg["signingCertificatePrivateKeyFile"] = "N/A"
        for i in range(self.root_key_record.number_of_certificates):
            key: Optional[ECC.EccKey] = None
            if i == self.root_key_record.used_root_cert:
                key = convert_to_ecc_key(self.root_key_record.root_public_key)
            else:
                if i < len(self.root_key_record.root_certs) and self.root_key_record.root_certs[i]:
                    key = convert_to_ecc_key(self.root_key_record.root_certs[i])
            if key:
                key_file_name = os.path.join(output_folder, f"rootCertificate{i}File.pub")
                misc.write_file(key.export_key(format="PEM"), key_file_name)
                cfg[f"rootCertificate{i}File"] = f"rootCertificate{i}File.pub"
            else:
                cfg[
                    f"rootCertificate{i}File"
                ] = "The public key is not possible reconstruct from the key hash"

        cfg["mainRootCertId"] = self.root_key_record.used_root_cert
        if self.isk_certificate and self.root_key_record.ca_flag == 0:
            cfg["useIsk"] = True
            assert self.isk_certificate.isk_cert
            key = self.isk_certificate.isk_cert
            key_file_name = os.path.join(output_folder, "signingCertificateFile.pub")
            misc.write_file(key.export_key(format="PEM"), key_file_name)
            cfg["signingCertificateFile"] = "signingCertificateFile.pub"
            cfg["signingCertificateConstraint"] = self.isk_certificate.constraints
            if self.isk_certificate.user_data:
                key_file_name = os.path.join(output_folder, "isk_user_data.bin")
                misc.write_file(self.isk_certificate.user_data, key_file_name, mode="wb")
                cfg["signCertData"] = "isk_user_data.bin"

        else:
            cfg["useIsk"] = False

        return cfg

    def create_config(self, data_path: str) -> str:
        """Create configuration of the Certification block Image.

        :param data_path: Path to store the data files of configuration.
        :return: Configuration in string.
        """
        cfg = self.get_config(data_path)
        val_schemas = CertBlockV31.get_validation_schemas()

        yaml_data = CommentedConfig(
            main_title=(
                "Certification block v3.1 recreated configuration from :"
                f"{datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}."
            ),
            schemas=val_schemas,
            values=cfg,
        ).export_to_yaml()
        return yaml_data


def get_main_cert_index(config: Dict[str, Any], search_paths: Optional[List[str]] = None) -> int:
    """Gets main certificate index from configuration.

    :param config: Input standard configuration.
    :param search_paths: List of paths where to search for the file, defaults to None
    :return: Instance of CertBlockV2
    :raises SPSDKError: If invalid configuration is provided.
    :raises SPSDKError: If correct certificate could not be identified.
    :raises SPSDKValueError: If certificate is not of correct type.
    """
    root_cert_id = config.get("mainRootCertId")
    cert_chain_id = config.get("mainCertChainId")
    if root_cert_id is not None and cert_chain_id is not None and root_cert_id != cert_chain_id:
        raise SPSDKError(
            "The mainRootCertId and mainRootCertId are specified and have different values."
        )
    found_cert_id = find_main_cert_index(config=config, search_paths=search_paths)
    if root_cert_id is None and cert_chain_id is None:
        if found_cert_id is not None:
            return found_cert_id
        else:
            raise SPSDKError("Certificate could not be found")
    # root_cert_id may be 0 which is falsy value, therefore 'or' cannot be used
    cert_id = root_cert_id if root_cert_id is not None else cert_chain_id
    try:
        cert_id = int(cert_id)  # type: ignore[arg-type]
    except ValueError as exc:
        raise SPSDKValueError(f"A certificate index is not a number: {cert_id}") from exc
    if found_cert_id is not None and found_cert_id != cert_id:
        logger.warning("Defined certificate does not match the private key.")
    return cert_id


def find_main_cert_index(
    config: Dict[str, Any], search_paths: Optional[List[str]] = None
) -> Optional[int]:
    """Go through all certificates and find the index matching to private key.

    :param config: Configuration to be searched.
    :param search_paths: List of paths where to search for the file, defaults to None
    :return: List of root certificates.
    """
    try:
        signature_provider = get_signature_provider(
            sp_cfg=config.get("signProvider"),
            local_file_key=config.get("mainCertPrivateKeyFile"),
            search_paths=search_paths,
        )
    except SPSDKError as exc:
        logger.debug(f"A signature provider could not be created: {exc}")
        return None

    root_certificates = find_root_certificates(config)
    public_keys = []
    for root_crt_file in root_certificates:
        try:
            public_key = loaders.extract_public_key(root_crt_file, search_paths=search_paths)
            public_keys.append(public_key)
        except SPSDKError:
            continue
    try:
        idx = get_matching_key_id(public_keys, signature_provider)
        return idx
    except ValueError:
        return None


def find_root_certificates(config: Dict[str, Any]) -> List[str]:
    """Find all root certificates in configuration.

    :param config: Configuration to be searched.
    :raises SPSDKError: If invalid configuration is provided.
    :return: List of root certificates.
    """
    root_certificates_loaded: List[Optional[str]] = [
        config.get(f"rootCertificate{idx}File") for idx in range(4)
    ]
    # filter out None and empty values
    root_certificates = list(filter(None, root_certificates_loaded))
    for org, filtered in zip(root_certificates_loaded, root_certificates):
        if org != filtered:
            raise SPSDKError("There are gaps in rootCertificateXFile definition")
    return root_certificates
